import balderhub.heart.lib.scenario_features
from balderhub.ant.lib.scenario_features import AntplusControllerHrmFeature


class BpmValueReaderFeature(balderhub.heart.lib.scenario_features.BpmValueReaderFeature):
    """
    Setup Level feature implementation of the ``balderhub.heart.lib.scenario_features.BpmValueReaderFeature`` while
    reading the values over ANT
    """
    ant_controller = AntplusControllerHrmFeature()

    def read_last_bpm_value(self):
        if not self.ant_controller.channel_is_active:
            self.ant_controller.open_channel()
        return self.ant_controller.wait_for_new_broadcast_message(timeout=1).computed_heart_rate



# TODO move this in a general available namespace??
from enum import Enum
from typing import Union

from .antplus_device_config import AntplusDeviceConfig
from ..utils import pages


class AntplusHrmDeviceConfig(AntplusDeviceConfig):
    """
    More specific scenario-level device config for devices that implement the HRM profile.
    """

    class SupportedSpecVersion(Enum):
        """enum describing the supported spec version this feature supports"""
        V2_0 = 2.0
        V2_1 = 2.1
        V2_5 = 2.5

    @property
    def supported_spec_version(self) -> SupportedSpecVersion:
        """gives the current supported spec version the device should support"""
        return self.SupportedSpecVersion.V2_5

    @property
    def support_battery_voltage_messuring(self) -> bool:
        """
        :return: returns True if the device supports battery voltage messages within its BatteryLevel Page, otherwise
                 False
        """
        return True

    @property
    def manufacturer_id(self):
        """
        :return: the expected manufacturer ID
        """
        raise NotImplementedError()

    @property
    def serial_number(self):
        """
        :return: the expected serial number of the device
        """
        raise NotImplementedError()

    @property
    def hardware_version(self):
        """
        :return: the expected hardware version of the device
        """
        raise NotImplementedError()

    @property
    def software_version(self):
        """
        :return: the expected software version of the device
        """
        raise NotImplementedError()

    @property
    def model_number(self):
        """
        :return: the expected model number of the device
        """
        raise NotImplementedError()

    @property
    def expected_main_page(
            self
    ) -> Union[type[pages.hrm.Hrm0DefaultDataPage], type[pages.hrm.Hrm4PreviousHeartBeatEventTimePage]]:
        """
        :return: returns the expected MAIN page
        """
        # TODO adjust to sports modes!
        return pages.hrm.Hrm4PreviousHeartBeatEventTimePage

    @property
    def expected_background_pages(
            self
    ) -> list[type[pages.hrm.BaseHrmPage]]:
        """
        :return: returns a list of all pages that are expected to be a BACKGROUND page
        """
        return [
            pages.hrm.Hrm2ManufacturerInformationPage,
            pages.hrm.Hrm3ProductInformationPage,
            pages.hrm.Hrm6CapabilitiesPage,
            pages.hrm.Hrm7BatteryStatusPage,
        ]

    @property
    def manual_request_possible_for(self) -> list[type[pages.hrm.BaseHrmPage]]:
        """
        :return: returns a list of all pages a manual Page-Request (PAGE ID 70) can be done with
        """
        return [
            pages.hrm.Hrm2ManufacturerInformationPage,
            pages.hrm.Hrm3ProductInformationPage,
            pages.hrm.Hrm7BatteryStatusPage,
        ]

    @property
    def manual_request_redirect_ack_as_broadcast(self) -> bool:
        """
        :return: returns True if it is expected that all requested Pages as ACK response are redirected as BROADCAST,
                 False otherwise
        """
        return False

    def get_expected_battery_state_for_level(self, battery_level: int) -> int:
        """
        Converts the given BatteryLevel into the expected Battery-State that is returned withing
        :class:`Hrm7BatteryStatusPage`

        :param battery_level: the battery level to get the expected Battery-State for
        :return: the battery state as integer
        """
        if battery_level == 0:
            return 0x05  # Critical
        if battery_level < 20:
            return 0x04  # Low
        if battery_level < 80:
            return 0x03  # OK
        if battery_level < 90:
            return 0x02 # Good
        return 0x01 # New



from __future__ import annotations

import struct
from ..base_antplus_page import BaseAntplusPage


class Common70RequestDataPage(BaseAntplusPage):
    """
    Common data page 70 to request a specific data page from the other device
    """
    PAGE_ID = 70

    _STRUCT_DATA_FORMAT = '<BBBBBBBB'

    @classmethod
    def create(
            cls,
            requested_transmission_response: int,  # TODO use bit field?
            requested_page_no: int,
            command_type: int,
    ) -> Common70RequestDataPage:
        """
        Creates a new message type with the provided field values.

        :param requested_transmission_response: the requested transmission response byte
        :param requested_page_no: the page number to request
        :param command_type: the command type to send
        :return: the ready to send message
        """
        reserved = 0xFF

        data = struct.pack(
            cls._STRUCT_DATA_FORMAT,
            cls.PAGE_ID,
            reserved,
            reserved,
            reserved,
            reserved,
            requested_transmission_response,
            requested_page_no,
            command_type,
        )
        return cls(data)

    @property
    def descriptor_byte1(self):
        """
        :return: returns the descriptor byte 1
        """
        return self._raw_data[3]

    @property
    def descriptor_byte2(self):
        """
        :return: returns the descriptor byte 2
        """
        return self._raw_data[4]

    @property
    def requested_transmission_response(self) -> int:
        """
        :return: returns the requested transmission response
        """
        return self._raw_data[5]

    @property
    def requested_no_of_times(self) -> int:
        """
        :return: returns the number of times, the device should respond with the requested page
        """
        return self.requested_transmission_response & ~(1 << 7)

    @property
    def requested_ack_response(self) -> bool:
        """
        :return: returns True if the page object asks for a response as ACK message instead of BROADCAST message
        """
        return bool(self.requested_transmission_response & (1 << 7))

    @property
    def requested_page_no(self) -> int:
        """
        :return: returns the requested page number
        """
        return self._raw_data[6]

    @property
    def command_type(self) -> int:
        """
        :return: returns the requested command type
        """
        return self._raw_data[7]



from .openant_manager_feature import OpenantManagerFeature
from .openant_plus_controller_hrm_feature import OpenantPlusControllerHrmFeature

__all__ = [
    'OpenantManagerFeature',
    'OpenantPlusControllerHrmFeature'
]



from .antplus_test_criteria_config import AntplusTestCriteriaConfig


class AntplusHrmTestCriteriaConfig(AntplusTestCriteriaConfig):
    """
    Heart-Rate-Monitor specific ANT+ test criteria configuration feature
    """

    def get_allowed_min_max_rr_value_for(self, expected_rr_value_sec: float) -> tuple[float, float]:
        """
        This method returns the allowed min and max value for a given RR-Value within a heart-rate monitor profile
        :param expected_rr_value_sec: the expected value of the RR-Value
        :return: a tuple with the min and the max allowed values the real RR-Value is allowed to have
        """
        return expected_rr_value_sec * 0.9, expected_rr_value_sec * 1.1

    def get_allowed_min_max_bpm_value_for(self, expected_bpm_value: int) -> tuple[int, int]:
        """
        This method returns the allowed min and max value for a given BPM value within a heart-rate monitor profile
        :param expected_bpm_value: the expected value of the BPM
        :return: a tuple with the min and the max allowed values the real BPM is allowed to have
        """
        return max(0, expected_bpm_value - 1), expected_bpm_value + 1



from __future__ import annotations
from .base_hrm_page import BaseHrmPage
from ...page_message_collection import PageMessageCollection


class Hrm9DeviceInformationPage(BaseHrmPage):
    """
    This is the ninth data page in the HRM profile that holds some device information within the page-specific
    bytes.
    """
    PAGE_ID = 9
    _STRUCT_DATA_FORMAT = '<BBBBHBB'

    @property
    def heart_beat_event_type(self) -> int:
        """
        :return: returns the heart beat event type (0: measured; 1: computed)
        """
        return self._raw_data[1] & ((1 << 1) | (1 << 0))

    @classmethod
    def validate_messages(
            cls,
            of_msg_collection: PageMessageCollection
    ) -> None:
        """
        Method to validate the correctness of messages from this type

        :param of_msg_collection: the message collection that should be validated (will automatically be filtered for
                                  messages from :class:`Hrm9DeviceInformationPage`)
        """
        msgs_of_interest = of_msg_collection.filter_by_type(page_type=cls)

        for msg in msgs_of_interest:
            assert (msg.raw_data[1] & 0xFC) == 0xFC, \
                (f"reserved bits in data byte 1 are not 0xFC (full byte is {hex(msg.raw_data[1])}) "
                 f"for message {msg}")
            assert msg.raw_data[2] == 0xFF, \
                f"reserved byte 2 is not 0xFF, is {hex(msg.raw_data[2])} for message {msg}"
            assert msg.raw_data[3] == 0xFF, \
                f"reserved byte 2 is not 0xFF, is {hex(msg.raw_data[3])} for message {msg}"



from __future__ import annotations
from typing import Union

from .base_hrm_page import BaseHrmPage


class Hrm6CapabilitiesPage(BaseHrmPage):
    """
    This is the sixth data page in the HRM profile that holds information about the capabilities within the
    page-specific bytes.
    """
    PAGE_ID = 6
    _STRUCT_DATA_FORMAT = '<BBBBHBB'

    @property
    def feature_supported_byte(self) -> int:
        """
        :return: returns the full byte with the information about which features are supported
        """
        return self._raw_data[2]

    @property
    def feature_enabled_byte(self) -> int:
        """
        :return: returns the full byte with the information about which features are enabled
        """
        return self._raw_data[3]

    def __get_value_for_bit_at_idx(self, idx) -> Union[int, None]:
        if not self.feature_supported_byte & (1 << idx):
            return None
        return bool(self.feature_enabled_byte * (1 << idx))

    @property
    def extended_running_feature_enabled(self) -> Union[bool, None]:
        """
        :return: returns True if the extended-running-feature is enabled, False if it is disabled and None if it is not
                 supported
        """
        return self.__get_value_for_bit_at_idx(0)

    @property
    def extended_cycling_feature_enabled(self) -> Union[bool, None]:
        """
        :return: returns True if the extended-cycling-feature is enabled, False if it is disabled and None if it is not
                 supported
        """
        return self.__get_value_for_bit_at_idx(1)

    @property
    def extended_swimming_feature_enabled(self) -> Union[bool, None]:
        """
        :return: returns True if the extended-swimming-feature is enabled, False if it is disabled and None if it is not
                 supported
        """
        return self.__get_value_for_bit_at_idx(2)

    @property
    def manufacturer_specific_feature_bit6_enabled(self) -> Union[bool, None]:
        """
        :return: returns True if the manufacturer-specific-bit6 is enabled, False if it is disabled and None if it is
                 not supported
        """
        return self.__get_value_for_bit_at_idx(6)

    @property
    def manufacturer_specific_feature_bit7_enabled(self) -> Union[bool, None]:
        """
        :return: returns True if the manufacturer-specific-bit7 is enabled, False if it is disabled and None if it is
                 not supported
        """
        return self.__get_value_for_bit_at_idx(7)



import logging
import math
import time

import balder
from balder.connections import DCPowerConnection

from balderhub.ant.lib.utils import pages
import balderhub.battery.lib.scenario_features
from balderhub.heart.lib.scenario_features import HeartBeatFeature, StrapDockingFeature

from .base_hrm_scenario import BaseHrmScenario
from ...lib.utils.support import filter_hrm_messages_by_toggle_bit_change

logger = logging.getLogger(__name__)


# TODO split up this scenario into single components to make sure that subset of tests can run even without test hw
class ScenarioHrmDeviceProfileFullTransmissionPattern(BaseHrmScenario):
    """Test scenario that observes a full ANT transmission pattern time frame and then validates requirements defined
    in the HRM device profile"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.transmission_pattern_sequence_count < 2:
            raise ValueError('the sequence need to be at least 2 times - some test rely on that')

    COUNT_OF_MAX_POSSIBLE_BACKGROUND_PAGES = 6

    DO_SEQUENCE_WITH_HEART_RATE = 60

    DO_WITH_BATTERY_LEVEL = 1.0

    class Heart(balder.Device):
        """heart beat simulating device"""
        heart = HeartBeatFeature()

    @balder.connect('HeartRateSensor', over_connection=DCPowerConnection)
    class BatterySimulator(balder.Device):
        """device manipulating the battery voltage"""
        sim = balderhub.battery.lib.scenario_features.RemovableBatterySimFeature()

    @balder.connect(Heart, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateSensor(BaseHrmScenario.HeartRateSensor):
        """device detecting the row heart rate"""
        # _is_cnn = IsConnected()  # TODO
        strap = StrapDockingFeature()

    @balder.connect(HeartRateSensor, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateHost(BaseHrmScenario.HeartRateHost):
        """device receiving the heart rate data"""

    @property
    def transmission_pattern_duration_sec(self):
        """expected seconds to receive a full transmission pattern"""
        # TODO maybe also considering overflow test of heart beat?
        # TODO maybe also considering overflow of Heart Beat Event Time? - just use max time of all
        return (64 + 4) * self.COUNT_OF_MAX_POSSIBLE_BACKGROUND_PAGES * 8070/32768

    @property
    def transmission_pattern_sequence_count(self):
        """amount of transmission pattern sequences that should be waited for receiving the initial data"""
        return 2

    @property
    def total_observing_time(self):
        """
        :return: returns the total observing time calculated based on
                 :meth:`ScenarioHrmDeviceProfileFullTransmissionPattern.transmission_pattern_duration_sec` multiplied
                 with :meth:`ScenarioHrmDeviceProfileFullTransmissionPattern.transmission_pattern_sequence_count`
        """
        total_time =  self.transmission_pattern_duration_sec * self.transmission_pattern_sequence_count
        total_time += 2 # wait a little bit longer
        return total_time

    @property
    def min_expected_heart_beats(self):
        """
        :return: returns the minimum expected heart beats according to the observation time given in
                 :meth:`ScenarioHrmDeviceProfileFullTransmissionPattern.total_observing_time` divided by heart beat
                 frequency
        """
        return math.floor(self.total_observing_time / (self.DO_SEQUENCE_WITH_HEART_RATE / 60) )

    @balder.fixture('variation')
    def heart_beat_established(self):
        """make sure that heart beat is established, before entering the variation"""
        yield from self.Heart.heart.fixt_make_sure_heart_beat_established(with_bpm=60, restore_entry_state=True)

    @balder.fixture('variation')
    def chest_strap_attached(self, heart_beat_established):  # pylint: disable=unused-argument
        """make sure that chest strap is attached, before entering the variation"""
        yield from self.HeartRateSensor.strap.fixt_make_sure_to_be_attached(restore_entry_state=True)

    @balder.fixture('variation')
    def device_powered_on(self, heart_beat_established, chest_strap_attached):  # pylint: disable=unused-argument
        """make sure that device is powered on, before entering the variation"""
        yield from self.BatterySimulator.sim.fixt_make_sure_device_is_powered_on(
            with_level=self.DO_WITH_BATTERY_LEVEL,
            restore_entry_state=True
        )

    @balder.fixture('variation')
    def ant_is_disconnected(self, device_powered_on):  # pylint: disable=unused-argument
        """make sure that ANT is disconnected, before entering the variation"""
        yield from self.HeartRateHost.controller.fixt_make_sure_ant_channel_is_closed()

    @balder.fixture('variation')
    def run_transmission_pattern_session(self, ant_is_disconnected):  # pylint: disable=unused-argument
        """
        fixture that runs the recoding session - during this fixture the normal transmission pattern is recorded,
        which will be analyzed by the tests within this scenario
        """
        logger.info(f'set heart beat to {self.DO_SEQUENCE_WITH_HEART_RATE}')
        self.Heart.heart.start(self.DO_SEQUENCE_WITH_HEART_RATE)

        logger.info('connect with ANT device')
        self.HeartRateHost.controller.open_channel()

        time_to_wait = self.total_observing_time

        logger.info(f'now wait for {time_to_wait:.2f} seconds to make sure that we receive '
                    f'{self.transmission_pattern_sequence_count} full transmission patterns')
        time.sleep(time_to_wait)
        logger.info('close ANT device channel')
        self.HeartRateHost.controller.close_channel()

        logger.info('stop heart beat')
        self.Heart.heart.stop()

    def test_general_profile_consistency(self):
        """
        This test executed the profile validation method
        :meth:`AntplusControllerHrmFeature.validate_profile_consistency` that validates all general valid conditions of
        the profile.
        """
        validation_report = self.HeartRateHost.controller.get_profile_consistency_validation_report()
        for cur_descr, cur_result in validation_report.items():
            if cur_result is None:
                logger.info(f"VALIDATION SUCCESSFULLY: {cur_descr}")
            else:
                logger.error(f"VALIDATION FAILED:      {cur_descr}: {cur_result}")
        errors_only = [(k, v) for k, v in validation_report.items() if v is not None]
        assert len(errors_only) == 0, ("detect errors within the profile: \n" +
                                       '\n'.join([f"- {k}: ERROR MESSAGE `{v}`" for k, v in errors_only]) + '\n')

    def test_validate_heart_beat_counts(self):
        """
        This test reads all messages and makes sure that there is no beat loss.
        """
        messages = self.HeartRateHost.controller.received_broadcast_messages

        last_beat_count = messages[0].heart_beat_count
        for idx, cur_msg in enumerate(messages):

            if last_beat_count == cur_msg.heart_beat_count:
                continue

            normalized_cur_beat_count = \
                cur_msg.heart_beat_count \
                    if last_beat_count < cur_msg.heart_beat_count \
                    else (cur_msg.heart_beat_count + 0x100)

            assert normalized_cur_beat_count == last_beat_count + 1, \
                (f"received unexpected beat count {cur_msg.heart_beat_count} (beat before was {last_beat_count}) "
                 f"in message {cur_msg} at index {idx}")

            last_beat_count = cur_msg.heart_beat_count

    def test_validate_heart_beat_event_time(self):
        """
        This test reads the beat time of all events and check that these values have a exact diff-time of the set heart
        rate.
        """

        messages = self.HeartRateHost.controller.received_broadcast_messages

        assert len(messages) > 0, "did not receive any messages"

        expected_diff_time_sec =  60 / self.DO_SEQUENCE_WITH_HEART_RATE
        allowed_min_diff_time_sec, allowed_max_diff_time_sec = \
            self.HeartRateSensor.test_criteria.get_allowed_min_max_rr_value_for(expected_diff_time_sec)
        allowed_min_diff_time = int(allowed_min_diff_time_sec * 1024)
        allowed_max_diff_time = int(allowed_max_diff_time_sec * 1024)

        logger.info('calculate normalized heat beat event time')
        skipped_beat_counts = 0
        last_beat_msg_idx = 0
        for idx, cur_message in enumerate(messages):
            last_beat_msg = messages[last_beat_msg_idx]
            # handle overflow if necessary
            cur_message_beat_count = cur_message.heart_beat_count \
                if cur_message.heart_beat_count > last_beat_msg.heart_beat_count \
                else (cur_message.heart_beat_count + 0x100)

            if last_beat_msg.heart_beat_count == cur_message.heart_beat_count:
                # same beat -> make sure that time is still the same
                assert last_beat_msg.heart_beat_event_time == cur_message.heart_beat_event_time
                continue

            # beat is different

            if skipped_beat_counts < self.HeartRateSensor.test_criteria.first_number_of_beats_to_skip:
                logger.debug(f'skip check between beat {last_beat_msg.heart_beat_count} (idx={last_beat_msg_idx}) '
                             f'and {cur_message.heart_beat_count} (idx={idx}) because test is configured to skip the '
                             f'first {self.HeartRateSensor.test_criteria.first_number_of_beats_to_skip} beats')
                skipped_beat_counts += 1
            else:
                # -> check that it is exactly one higher
                assert cur_message_beat_count == last_beat_msg.heart_beat_count + 1, \
                    (f"unexpected beat count {cur_message.heart_beat_count} of msg at idx {idx} "
                     f"(beat count before was {last_beat_msg.heart_beat_count}) )")

                cur_message_event_time = (cur_message.heart_beat_event_time + 0x10000) \
                    if cur_message.heart_beat_event_time < last_beat_msg.heart_beat_event_time \
                    else cur_message.heart_beat_event_time

                cur_message_diff_time = cur_message_event_time - last_beat_msg.heart_beat_event_time

                assert allowed_min_diff_time <= cur_message_diff_time <= allowed_max_diff_time, \
                    (f"difference detected between heart beat {last_beat_msg.heart_beat_count} "
                     f"(idx: {last_beat_msg_idx}) and heart beat {cur_message.heart_beat_count} (idx: {idx}): "
                     f"received: {cur_message_diff_time} (expected value between {allowed_min_diff_time} and "
                     f"{allowed_max_diff_time} for configured {self.DO_SEQUENCE_WITH_HEART_RATE} BPM)")
            last_beat_msg_idx = idx

    def test_main_page_0_default(self):
        """
        This test validates the content of the main page 0, if it is expected that this page is the main page.
        """
        page_type = pages.hrm.Hrm0DefaultDataPage

        if self.HeartRateSensor.ant_config.expected_main_page == page_type:
            logger.info(f'make sure page 0 is the active main page '
                        f'(expected behavior according `{self.HeartRateSensor.ant_config.__class__}`)')
            existing_main_pages = self.HeartRateHost.controller.determine_main_pages()
            assert page_type in existing_main_pages, \
                (f"did not find expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_main_pages}")
            # nothing more to test, because data is empty
            all_msgs = self.HeartRateHost.controller.received_broadcast_messages
            msgs_of_interest = all_msgs.filter_by_type(page_type=page_type)

            for cur_msg in msgs_of_interest:
                msg_idx = all_msgs.index(cur_msg)
                assert cur_msg.raw_data[1] == 0xFF, \
                    f"reserved byte 1 is not 0xFF, is {hex(cur_msg.raw_data[1])} for message at index {msg_idx}"
                assert cur_msg.raw_data[2] == 0xFF, \
                    f"reserved byte 2 is not 0xFF, is {hex(cur_msg.raw_data[2])} for message at index {msg_idx}"
                assert cur_msg.raw_data[3] == 0xFF, \
                    f"reserved byte 2 is not 0xFF, is {hex(cur_msg.raw_data[3])} for message at index {msg_idx}"
        else:
            logger.info(f'make sure page 0 is NOT an active main page '
                        f'(expected behavior according `{self.HeartRateSensor.ant_config.__class__}`)')
            existing_main_pages = self.HeartRateHost.controller.determine_main_pages()
            existing_background_pages = self.HeartRateHost.controller.determine_background_pages()

            assert page_type not in existing_main_pages, \
                (f"found not-expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_main_pages}")
            assert page_type not in existing_background_pages, \
                (f"found not-expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_background_pages}")

    def test_background_page_1_operating_time(self):
        """
        This test validates the content of the background page 1, which was sent during the session, if it was
        expected, that this page is an active background page.
        """
        page_type = pages.hrm.Hrm1CumulativeOperationTimePage

        if page_type in self.HeartRateSensor.ant_config.expected_background_pages:
            logger.info(f'make sure page 1 is a background page '
                        f'(expected behavior according `{self.HeartRateSensor.ant_config.__class__}`)')
            existing_background_pages = self.HeartRateHost.controller.determine_background_pages()
            assert page_type in existing_background_pages, \
                (f"did not find expected background page `{page_type}` within determined BACKGROUND "
                 f"pages {existing_background_pages}")

            # todo check that we have received at least two cycles (necessary - check in __init__()!)
            page_type.validate_messages(
                of_msg_collection=self.HeartRateHost.controller.received_broadcast_messages,
            )
        else:
            logger.info(f'make sure page 1 is not a background page '
                        f'(expected behavior according `{self.HeartRateSensor.ant_config.__class__}`)')
            existing_main_pages = self.HeartRateHost.controller.determine_main_pages()
            existing_background_pages = self.HeartRateHost.controller.determine_background_pages()

            assert page_type not in existing_main_pages, \
                (f"found not-expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_background_pages}")
            assert page_type not in existing_background_pages, \
                (f"found not-expected main page `{page_type}` within determined BACKGROUND "
                 f"pages {existing_background_pages}")

    def test_background_page_2_manufacturer(self):
        """
        This test validates the content of the background page 2, which was sent during the session, if it was
        expected, that this page is an active background page.
        """
        page_type = pages.hrm.Hrm2ManufacturerInformationPage

        existing_background_pages = self.HeartRateHost.controller.determine_background_pages()
        assert page_type in existing_background_pages, \
            f"page type {page_type} not found in background page list: `{existing_background_pages}`"

        # TODO maybe use coherent messages only
        msg = filter_hrm_messages_by_toggle_bit_change(
            self.HeartRateHost.controller.received_broadcast_messages
        ).filter_by_type(page_type=page_type)

        assert len(msg) >= self.transmission_pattern_sequence_count, \
            (f"expect more than {self.transmission_pattern_sequence_count} messages of type {page_type}, "
             f"because the transmission pattern sequence count if {self.transmission_pattern_sequence_count}, "
             f"but just detect {len(msg)} independent messages")

        self.HeartRateHost.controller.validate_page_2_manufacturer()

    def test_background_page_3_product(self):
        """
        This test validates the content of the background page 3, which was sent during the session, if it was
        expected, that this page is an active background page.
        """
        page_type = pages.hrm.Hrm3ProductInformationPage

        existing_background_pages = self.HeartRateHost.controller.determine_background_pages()
        assert page_type in existing_background_pages, \
            f"page type {page_type} not found in background page list: `{existing_background_pages}`"

        # TODO maybe use coherent messages only
        msg = filter_hrm_messages_by_toggle_bit_change(
            self.HeartRateHost.controller.received_broadcast_messages
        ).filter_by_type(page_type=page_type)

        assert len(msg) >= self.transmission_pattern_sequence_count, \
            (f"expect more than {self.transmission_pattern_sequence_count} messages of type {page_type}, "
             f"because the transmission pattern sequence count if {self.transmission_pattern_sequence_count}, "
             f"but just detect {len(msg)} independent messages")

        self.HeartRateHost.controller.validate_page_3_product()

    def test_main_page_4_previous_beat(self):
        """
        This test validates the content of the main page 4, if it is expected that this page is the main page.
        """
        page_type = pages.hrm.Hrm4PreviousHeartBeatEventTimePage

        if self.HeartRateSensor.ant_config.expected_main_page == page_type:
            logger.info('make sure that page 4 is a normal main page')
            existing_main_pages = self.HeartRateHost.controller.determine_main_pages()
            assert page_type in existing_main_pages, \
                (f"did not find expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_main_pages}")
            # nothing more to test, because data is empty
            logger.info('validate that all previous heart-beat-event times are valid')

            all_msgs = self.HeartRateHost.controller.received_broadcast_messages
            msgs_of_interest = filter_hrm_messages_by_toggle_bit_change(all_msgs).filter_by_type(page_type=page_type)

            def get_last_different_heart_beat_event_time(for_idx):
                cur_event_time = all_msgs[for_idx].heart_beat_event_time
                for inner_idx in reversed(range(0, for_idx)):
                    if all_msgs[inner_idx].heart_beat_event_time != cur_event_time:
                        return all_msgs[inner_idx].heart_beat_event_time
                raise ValueError('no other value found')

            non_first_beat_msgs = [msg for msg in all_msgs if msg.heart_beat_count != all_msgs[0].heart_beat_count]
            assert len(non_first_beat_msgs) > 0, "did not found messages with a heart beat"
            first_beat_idx = all_msgs.index(non_first_beat_msgs[0])
            for idx, msg in enumerate(all_msgs):
                # iterating over all messages, because the current event time is always valid, but
                # validating the previous-heart-beat-event-time only when toggle-bit changed (and instance of this type)
                if msg in msgs_of_interest and idx >= first_beat_idx:
                    last_heart_beat_event_time = get_last_different_heart_beat_event_time(idx)
                    assert last_heart_beat_event_time == msg.previous_heart_beat_event_time_raw, \
                        (f"received invalid previous-heart-beat-event-time "
                         f"{msg.previous_heart_beat_event_time_raw} at index {idx} "
                         f"(but last transmitted event time was {last_heart_beat_event_time})")

            logger.info(f'validated {len(msgs_of_interest)} messages -> all of them are valid')

        else:
            logger.info('make sure page 4 is a normal main page (and with that also no background page)')
            existing_main_pages = self.HeartRateHost.controller.determine_main_pages()
            existing_background_pages = self.HeartRateHost.controller.determine_background_pages()

            assert page_type not in existing_main_pages, \
                (f"found not-expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_main_pages}")
            assert page_type not in existing_background_pages, \
                (f"found not-expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_background_pages}")

    # TODO
    #def test_main_page_5_swim_interval_summary(self):
    #    raise NotImplementedError

    def test_background_page_6_capabilities(self):
        """
        This test validates the content of the background page 6, which was sent during the session, if it was
        expected, that this page is an active background page.
        """
        # TODO reqirement of that in background page was optional before spec 2.5
        page_type = pages.hrm.Hrm6CapabilitiesPage

        supported_spec_version = self.HeartRateSensor.ant_config.supported_spec_version
        if supported_spec_version == self.HeartRateSensor.ant_config.SupportedSpecVersion.V2_5:
            assert page_type in self.HeartRateSensor.ant_config.expected_background_pages, \
                (f"your configuration violates the spec: {page_type} needs to be a background page - "
                 f"please adjust your `AntplusHrmDeviceConfig.expected_background_pages` setting")

        existing_background_pages = self.HeartRateHost.controller.determine_background_pages()
        if page_type in self.HeartRateSensor.ant_config.expected_background_pages:

            assert page_type in existing_background_pages, \
                f"page type {page_type} not found in background page list: `{existing_background_pages}`"

            all_msgs = self.HeartRateHost.controller.received_broadcast_messages
            msgs_of_interest = all_msgs.filter_by_type(page_type=page_type)
            for msg in msgs_of_interest:
                msg_idx = all_msgs.index(msg)
                assert msg.raw_data[1] == 0xFF, \
                    f"reserved byte 1 is not 0xFF, is {hex(msg.raw_data[2])} for message at index {msg_idx}"
                assert msg.raw_data[2] & ((1 << 4) | (1 << 5)) == 0x00, \
                    (f"reserved bits 4 and 5 of byte 2 are not 0, byte has value {hex(msg.raw_data[2])} for message "
                     f"at index {msg_idx}")
                assert msg.raw_data[3] & ((1 << 4) | (1 << 5)) == 0x00, \
                    (f"reserved bits 4 and 5 of byte 3 are not 0, byte has value {hex(msg.raw_data[3])} for message "
                     f"at index {msg_idx}")
                active_bits = msg.raw_data[2]
                assert (msg.raw_data[3] & ~active_bits) == 0x00, \
                    (f"some bits are active in ENABLED (byte 3: {hex(msg.raw_data[3])}) "
                     f"but 0 in SUPPORTED (byte 2: {hex(msg.raw_data[2])}) for message at index {msg_idx}")
        else:
            assert page_type not in existing_background_pages, \
                f"page type {page_type} unexpectedly found in background page list: `{existing_background_pages}`"

    def test_background_page_7_battery(self):
        """
        This test validates the content of the background page 7, which was sent during the session, if it was
        expected, that this page is an active background page.
        """
        page_type = pages.hrm.Hrm7BatteryStatusPage

        if page_type in self.HeartRateSensor.ant_config.expected_background_pages:
            logger.info(f'make sure page 7 is a background page (expected behavior according '
                        f'`{self.HeartRateSensor.ant_config.__class__}`)')
            existing_background_pages = self.HeartRateHost.controller.determine_background_pages()
            assert page_type in existing_background_pages, \
                (f"did not find expected background page `{page_type}` within determined BACKGROUND "
                 f"pages {existing_background_pages}")
            # nothing more to test, because data is empty
            all_msgs = self.HeartRateHost.controller.received_broadcast_messages

            expected_bat_voltage = \
                self.BatterySimulator.sim.discharge_characteristic.get_voltage_for(self.DO_WITH_BATTERY_LEVEL) \
                    if self.HeartRateSensor.ant_config.support_battery_voltage_messuring else None

            page_type.validate_messages(
                of_msg_collection=all_msgs,
                expected_battery_level=int(self.DO_WITH_BATTERY_LEVEL * 100),
                expected_battery_status=self.HeartRateSensor.ant_config.get_expected_battery_state_for_level(
                    int(self.DO_WITH_BATTERY_LEVEL * 100)
                ),
                expected_battery_voltage=expected_bat_voltage,
            )

        else:
            logger.info(f'make sure page 7 is not a background page '
                        f'(expected behavior according `{self.HeartRateSensor.ant_config.__class__}`)')
            existing_main_pages = self.HeartRateHost.controller.determine_main_pages()
            existing_background_pages = self.HeartRateHost.controller.determine_background_pages()

            assert page_type not in existing_main_pages, \
                (f"found not-expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_background_pages}")
            assert page_type not in existing_background_pages, \
                (f"found not-expected backend page `{page_type}` within determined BACKGROUND "
                 f"pages {existing_background_pages}")

    def test_background_page_9_device_info(self):
        """
        This test validates the content of the background page 9, which was sent during the session, if it was
        expected, that this page is an active background page.
        """
        # check that it is not within the transmission or if data is valid (depending on setting)
        page_type = pages.hrm.Hrm9DeviceInformationPage

        if page_type in self.HeartRateSensor.ant_config.expected_background_pages:
            logger.info(f'make sure page 9 is a background page '
                        f'(expected behavior according `{self.HeartRateSensor.ant_config.__class__}`)')
            existing_background_pages = self.HeartRateHost.controller.determine_background_pages()
            assert page_type in existing_background_pages, \
                (f"did not find expected background page `{page_type}` within determined BACKGROUND "
                 f"pages {existing_background_pages}")
            # nothing more to test, because data is empty
            page_type.validate_messages(
                of_msg_collection=self.HeartRateHost.controller.received_broadcast_messages,
            )
        else:
            logger.info(f'make sure page 9 is not a background page '
                        f'(expected behavior according `{self.HeartRateSensor.ant_config.__class__}`)')
            existing_main_pages = self.HeartRateHost.controller.determine_main_pages()
            existing_background_pages = self.HeartRateHost.controller.determine_background_pages()

            assert page_type not in existing_main_pages, \
                (f"found not-expected main page `{page_type}` within determined MAIN "
                 f"pages {existing_background_pages}")
            assert page_type not in existing_background_pages, \
                (f"found not-expected background page `{page_type}` within determined BACKGROUND "
                 f"pages {existing_background_pages}")

    def test_no_other_background_pages_exists(self):
        """test that validates that every expected background page is within the recorded pages"""
        for existing_background_page in self.HeartRateHost.controller.determine_background_pages():
            assert existing_background_page in self.HeartRateSensor.ant_config.expected_background_pages, \
                (f"detect a background page `{existing_background_page}` in data that is not existing in the "
                 f"expected background pages: {self.HeartRateSensor.ant_config.expected_background_pages}")

    def test_no_other_main_pages_exists(self):
        """test that validates that no other main pages, then the expected one, are within the recorded pages"""

        existing_main_pages = self.HeartRateHost.controller.determine_main_pages()
        assert len(existing_main_pages) == 1, f"detect more than one main pages `{existing_main_pages}`"
        assert existing_main_pages[0] == self.HeartRateSensor.ant_config.expected_main_page, \
                (f"detect a unexpected main page `{existing_main_pages[0]}` in data that is not the expected one "
                 f"`{self.HeartRateSensor.ant_config.expected_main_page}`")



from __future__ import annotations

from .base_hrm_page import BaseHrmPage


class Hrm0DefaultDataPage(BaseHrmPage):
    """
    This is the default data page without any specific information in the page specific bytes.
    """
    PAGE_ID = 0
    _STRUCT_DATA_FORMAT = '<BBBBHBB'



import time

import balder
from balder.connections import DCPowerConnection

import balderhub.battery.lib.scenario_features
from balderhub.battery.lib.scenario_features import BatteryTestCriteriaConfig

from balderhub.ant.scenarios.hrm.base_hrm_scenario import BaseHrmScenario

from balderhub.ant.lib.utils import pages
from balderhub.heart.lib.scenario_features import HeartBeatFeature, StrapDockingFeature


class ScenarioHrmBatteryMeasuring(BaseHrmScenario):
    """
    Test Scenario validating the sending of the correct battery values when the battery level is changed
    """

    @balder.connect('HeartRateSensor', over_connection=DCPowerConnection)
    class BatterySimulator(balder.Device):
        """device manipulating the battery voltage"""
        sim = balderhub.battery.lib.scenario_features.RemovableBatterySimFeature()

    class Heart(balder.Device):
        """device simulating a heart beat"""
        heart = HeartBeatFeature()

    @balder.connect(Heart, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateSensor(BaseHrmScenario.HeartRateSensor):
        """device detecting the row heart rate"""
        strap = StrapDockingFeature()
        test_config = BatteryTestCriteriaConfig()

    @balder.connect(HeartRateSensor, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateHost(BaseHrmScenario.HeartRateHost):
        """device receiving the heart rate data"""

    @balder.fixture('variation')
    def disconnect_if_necessary(self):
        """fixture that makes sure that current sensor is disconnected"""
        yield from self.HeartRateHost.controller.fixt_make_sure_ant_channel_is_closed(restore_entry_state=True)

    @balder.fixture('variation')
    def heart_beat_established(self, disconnect_if_necessary):  # pylint: disable=unused-argument
        """make sure that heart beat is established, before entering the variation"""
        yield from self.Heart.heart.fixt_make_sure_heart_beat_established(with_bpm=60, restore_entry_state=True)

    @balder.fixture('variation')
    def chest_strap(self, heart_beat_established):  # pylint: disable=unused-argument
        """make sure that the chest strap is attached, before entering the variation"""
        yield from self.HeartRateSensor.strap.fixt_make_sure_to_be_attached(restore_entry_state=True)

    @balder.fixture('variation')
    def power_off_device(self, disconnect_if_necessary, heart_beat_established, chest_strap):  # pylint: disable=unused-argument
        """make sure that the device is powered off, before entering the variation"""
        yield from self.BatterySimulator.sim.fixt_make_sure_device_is_powered_off(restore_entry_state=True)

    @balder.fixture('testcase')
    def wait_for_reset(self, power_off_device):  # pylint: disable=unused-argument
        """wait a second before entering any testcase"""
        time.sleep(1)

    @balder.parametrize_by_feature(
        "battery_level", (HeartRateSensor, 'test_config', 'validation_with_battery_levels')
    )
    def test_check_different_measurements(self, battery_level):
        """
        Test that validates if the different battery levels are presented by the dut while changing them with the
        battery simulator.

        :param battery_level: feature paramized battery level (provided by
                              :meth:`BatteryTestCriteriaConfig.validation_with_battery_levels`)
        """
        self.BatterySimulator.sim.set_to(battery_level)
        expected_voltage = self.BatterySimulator.sim.discharge_characteristic.get_voltage_for(battery_level)
        try:
            self.BatterySimulator.sim.insert_battery()
            try:
                self.HeartRateHost.controller.open_channel()
                data = self.HeartRateHost.controller.wait_for_new_broadcast_message(
                    pages.hrm.Hrm7BatteryStatusPage, timeout=300
                )
                # todo add allowed deviation
                assert data.total_battery_voltage == 3.1, \
                    f"detect unexpected battery voltage {data.total_battery_voltage}V (expected {expected_voltage}V)"
            finally:
                self.HeartRateHost.controller.close_channel()
        finally:
            self.BatterySimulator.sim.remove_battery()



from collections import OrderedDict

import balder

from balderhub.ant.lib.utils import pages


class BaseAntplusDeviceProfile(balder.Feature):
    """
    Base class for an ANT+ Device Profile
    """

    ALL_PAGES = None

    @classmethod
    def get_existing_pages_for_profile(cls) -> OrderedDict[int, pages.BaseAntplusPage]:
        """
        :return: returns an ordered dictionary mapping with the Page-ID and the Page this profile supports
        """

        if cls.ALL_PAGES is None:
            raise ValueError(f'you need to define ALL_PAGES within your profile {cls.__name__}')

        result = []
        for page in sorted(cls.ALL_PAGES, key=lambda x: x.PAGE_ID):
            if page.PAGE_ID is None:
                raise ValueError('you need to define a PAGE_ID as soon as the page is used within a profile')
            result.append((page.PAGE_ID, page))
        return OrderedDict(result)



from __future__ import annotations
from .base_hrm_page import BaseHrmPage


class Hrm4PreviousHeartBeatEventTimePage(BaseHrmPage):
    """
    This is the fourth data page in the HRM profile that can also be used as a main page. It hold some manufacturer
    specific data and the previous heart beat event time within the page-specific bytes.
    """
    PAGE_ID = 4
    _STRUCT_DATA_FORMAT = '<BBBBHBB'

    @property
    def manufacturer_specific_byte(self) -> int:
        """
        :return: returns some manufacturer specific information
        """
        return self._raw_data[1]

    @property
    def previous_heart_beat_event_time_raw(self) -> int:
        """
        :return: returns the previous heart beat event time
        """
        return int.from_bytes(self.raw_data[2:4], 'little')



from .base_hrm_page import BaseHrmPage
from .hrm_0_default_data_page import Hrm0DefaultDataPage
from .hrm_1_cumulative_operation_time_page import Hrm1CumulativeOperationTimePage
from .hrm_2_manufacturer_information_page import Hrm2ManufacturerInformationPage
from .hrm_3_product_information_page import Hrm3ProductInformationPage
from .hrm_4_previous_heart_beat_event_time_page import Hrm4PreviousHeartBeatEventTimePage
from .hrm_5_swim_interval_summary_page import Hrm5SwimIntervalSummaryPage
from .hrm_6_capabilities_page import Hrm6CapabilitiesPage
from .hrm_7_battery_status_page import Hrm7BatteryStatusPage
from .hrm_9_device_information_page import Hrm9DeviceInformationPage
from .hrm_32_hr_feature_page import Hrm32HrFeaturePage

__all__ = [
    "BaseHrmPage",
    "Hrm0DefaultDataPage",
    "Hrm1CumulativeOperationTimePage",
    "Hrm2ManufacturerInformationPage",
    "Hrm3ProductInformationPage",
    "Hrm4PreviousHeartBeatEventTimePage",
    "Hrm5SwimIntervalSummaryPage",
    "Hrm6CapabilitiesPage",
    "Hrm7BatteryStatusPage",
    "Hrm9DeviceInformationPage",
    "Hrm32HrFeaturePage",
]



import balderhub.battery.lib.scenario_features
from balderhub.ant.lib.scenario_features import AntplusControllerHrmFeature


class DeviceActivityFeature(balderhub.battery.lib.scenario_features.DeviceActivityFeature):
    """Setup Level implementation for returning the activity of an ANT device by checking if it sends messages"""

    ant_controller = AntplusControllerHrmFeature()

    def is_active(self) -> bool:
        if not self.ant_controller.channel_is_active:
            self.ant_controller.open_channel()
        try:
            self.ant_controller.wait_for_new_broadcast_message(timeout=1)
            return True
        except TimeoutError:
            return False



from __future__ import annotations
from typing import Union, OrderedDict, Callable, Generator
import logging

import balder
from balderhub.ant.lib.scenario_features.antplus_device_config import AntplusDeviceConfig
from balderhub.ant.lib.scenario_features.base_antplus_device_profile import BaseAntplusDeviceProfile
from balderhub.ant.lib.utils.page_message_collection import PageMessageCollection
from balderhub.ant.lib.utils.pages import BaseAntplusPage

logger = logging.getLogger(__name__)


class AntplusControllerFeature(balder.Feature):
    """
    Base ANT+ Controller Feature that can be used for any Profile. It holds a inner VDevice that needs to define exactly
    one profile that is based on :class:`BaseAntplusDeviceProfile`.

    This scenario-level feature is used as base class for different type of controllers that are specified for a
    specific profile type.
    """

    class AntPlusDevice(balder.VDevice):
        """vdevice holding the ANT+ device profile"""
        config = AntplusDeviceConfig()
        profile = BaseAntplusDeviceProfile()

    class ValidationError(Exception):
        """error that is raised by internal profile validation functions"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._already_saved_broadcast_messages = PageMessageCollection()
        self._already_saved_ack_messages = PageMessageCollection()
        self._already_saved_burst_messages = PageMessageCollection()

    @property
    def validation_methods(self) -> OrderedDict[str, Callable[[], None]]:
        """
        :return: mapping of validation methods that can be used to check the full validity of all messages sent by a
                 ANT+ device as ordered dict (key is printable name and the callable as value)
        """
        return OrderedDict()

    @property
    def channel_type(self) -> int:  # TODO maybe use own enums
        """
        :return: the own channel type the controller should use
        """
        raise NotImplementedError

    @property
    def rf_channel_frequency(self) -> int:
        """
        :return: the own RF Channel the controller should use
        """
        raise NotImplementedError

    @property
    def transmission_type(self) -> int:  # TODO maybe use own enums
        """
        :return: the own transmission type the controller should use
        """
        raise NotImplementedError

    @property
    def device_type(self) -> int:  # TODO maybe use own enums
        """
        :return: the device type the controller should search for
        """
        raise NotImplementedError

    @property
    def channel_period(self) -> int:
        """
        :return: the channel period the controller should use
        """
        raise NotImplementedError

    @property
    def channel_is_active(self) -> bool:
        """
        :return: returns True if the channel is active, otherwise False
        """
        raise NotImplementedError

    @property
    def received_broadcast_messages(self) -> PageMessageCollection:
        """
        :return: returns the current available BROADCAST messages that has been received since the channel is active
        """
        return self._already_saved_broadcast_messages

    @property
    def received_ack_messages(self) -> PageMessageCollection:
        """
        :return: returns the current available ACK messages that has been received since the channel is active
        """
        return self._already_saved_ack_messages

    # TODO not supported yet
    #@property
    #def received_burst_messages(self) -> PageMessageCollection:
    #    return self._already_saved_burst_messages

    def open_channel(self) -> None:
        """
        Opens the channel
        """
        raise NotImplementedError()

    def close_channel(self) -> None:
        """
        Closes the Channel
        """
        raise NotImplementedError()

    def send_broadcast_message(self, message: BaseAntplusPage) -> None:
        """
        This method sends a given message as BROADCAST message within the open channel
        :param message: the message that should be sent
        """
        raise NotImplementedError()

    def send_ack_message(self, message: BaseAntplusPage) -> None:
        """
        This method sends a given message as ACK message within the open channel
        :param message: the message that should be sent
        """
        raise NotImplementedError()

    def wait_for_new_broadcast_message(
            self,
            of_page_type: Union[list[type[BaseAntplusPage]], type[BaseAntplusPage], None] = None,
            timeout: float = 10
    ) -> BaseAntplusPage:
        """
        This method waits for a new BROADCAST message to be received. All messages that has been received before
        entering this method are not considered.

        If no message is received within the given timeout, the method raises a ``TimeoutError``.

        :param of_page_type: if given, the only considers messages of this specific type
        :param timeout: the maximum time in seconds to wait for a new message
        :return: the new received message
        """
        raise NotImplementedError()

    def wait_for_new_ack_message(
            self,
            of_page_type: Union[list[type[BaseAntplusPage]], type[BaseAntplusPage], None] = None,
            timeout: float = 10
    ) -> BaseAntplusPage:
        """
        This method waits for a new ACK message to be received. All messages that has been received before
        entering this method are not considered.

        If no message is received within the given timeout, the method raises a ``TimeoutError``.

        :param of_page_type: if given, the only considers messages of this specific type
        :param timeout: the maximum time in seconds to wait for a new message
        :return: the new received message
        """
        raise NotImplementedError()

    def get_profile_consistency_validation_report(self) -> OrderedDict[str, Union[str, None]]:
        """
        This method is used to execute a set of validation_functions that makes sure that check that can be applied
        on the raw data of the profile.

        :return: an ordered dict with a description of the validation as key and the error string as value
                 (or None, if sub validation passed)
        """
        result = OrderedDict()
        for cur_descr, cur_callback in self.validation_methods.items():
            try:
                cur_callback()
                result[cur_descr] = None
            except self.ValidationError as exc:
                result[cur_descr] = exc.args[0]
            except Exception as exc:
                raise RuntimeError(f'unexpected error occurred in validation callback `{cur_callback}` '
                                   f'(use ValidationError if it was an validation error)') from exc
        return result

    def validate_profile_consistency(self) -> None:
        """
        This method is used to execute a set of validation_functions that makes sure that check that can be applied
        on the raw data of the profile.
        """
        report = self.get_profile_consistency_validation_report()
        errors = {k:v for k, v in report.items() if v is not None}
        if errors:
            raise ValueError('profile consistency has problems: \n' + '\n'.join(
                [f"- {k}: ERROR MESSAGE `{v}`" for k, v in errors.items()]
            ))

    def _fixt_teardown(self, channel_state_before: bool):
        if channel_state_before:
            if self.channel_is_active:
                logger.info('ANT+ Channel was opened before entering this fixture and it is still opened '
                            '-> do nothing')
            else:
                logger.info('ANT+ Channel was opened before entering this fixture, but it is closed now '
                            '-> open it again')
                self.open_channel()
        else:
            # channel was closed before this fixture
            if self.channel_is_active:
                logger.info('ANT+ Channel was closed before entering this fixture but it is opened now '
                            '-> close it again')
                self.close_channel()
            else:
                logger.info('ANT+ Channel was closed before entering this fixture and it is still closed '
                            '-> do nothing')

    def fixt_make_sure_ant_channel_is_opened(self, restore_entry_state: bool = True) -> Generator[None, None, None]:
        """
        Fixture that makes sure that the ANT Channel is opened. If it is already open when entering this fixture, it
        will do nothing. If it is closed, it will ensure that it is opened as soon as the construction part of this
        fixture is finished.

        .. note::

            You can use this fixture directly by using:

            .. code-block:: python

                @balder.fixture(...)
                def my_fixture(...):
                    yield from feat.fixt_make_sure_ant_channel_is_opened(...)

        :param restore_entry_state: True if the previous state (either open or closed) should be
                                    reestablished in the teardown part of this fixture
        :return: the ready to use fixture generator, but without any values
        """
        channel_was_open = False
        if self.channel_is_active:
            logger.info('ANT+ Channel is already active -> do nothing')
            channel_was_open = True
        else:
            logger.info('ANT+ Channel is inactive -> open it')
            self.open_channel()

        yield None

        if not restore_entry_state:
            return
        self._fixt_teardown(channel_state_before=channel_was_open)

    def fixt_make_sure_ant_channel_is_closed(self, restore_entry_state: bool = True) -> Generator[None, None, None]:
        """
        Fixture that makes sure that the ANT Channel is opened before finishing the construction part of this
        fixture.

        If `restore_entry_state` is True, the method will reestablish the state that was given before entering this
        fixture.

        .. note::

            You can use this fixture directly by using:

            .. code-block:: python

                @balder.fixture(...)
                def my_fixture(...):
                    yield from feat.fixt_make_sure_ant_channel_is_closed(...)

        :param restore_entry_state: True if the previous state (either connected or disconnected) should be
                                    reestablished in the teardown part of this fixture
        :return: the ready to use fixture generator, but without any values
        """
        channel_was_open = False
        if self.channel_is_active:
            logger.info('ANT+ Channel is active -> close it')
            channel_was_open = True
            self.close_channel()
        else:
            logger.info('ANT+ Channel is already inactive -> do nothing')

        yield None

        if not restore_entry_state:
            return
        self._fixt_teardown(channel_state_before=channel_was_open)



import copy


class BaseExtendedMetaLegacy:
    """
    Base class for metadata describing data given by the ANT interface in LEGACY-EXTENDED-DATA message format
    """
    # count of bytes the message has normally (without the flag byte)
    EXPECTED_BYTE_LENGTH = None
    BYTE_ORDER = 'little'

    def __init__(self, raw_data: bytes):
        if not isinstance(raw_data, bytes):
            raise TypeError(f'Expected bytes but got {type(raw_data)}')
        if len(raw_data) != self.EXPECTED_BYTE_LENGTH:
            raise ValueError(f'provided raw data needs to have a length of '
                             f'{self.EXPECTED_BYTE_LENGTH} bytes (is `{raw_data}`)')
        self._raw_data = raw_data

    @property
    def raw_data(self) -> bytes:
        """
        :return: returns the raw bytes that describing the information within this metadata object
        """
        return copy.copy(self._raw_data)



from .common_70_request_data_page import Common70RequestDataPage
from .common_76_mode_settings_page import Common76ModeSettingsPage


__all__ = [
    'Common70RequestDataPage',
    'Common76ModeSettingsPage'
]



import array
import logging
import queue
import time
from typing import Union, Literal

from openant.base.message import Message
from openant.easy.channel import Channel

from .openant_manager_feature import OpenantManagerFeature
from ..scenario_features.antplus_controller_hrm_feature import AntplusControllerHrmFeature
from ..utils.page_message_collection import PageMessageCollection
from ..utils.pages import BaseAntplusPage, BaseReceivedAntplusPage
from ..utils.extended_meta.extended_meta_flagged_channel_id import ExtendedMetaFlaggedChannelId
from ..utils.extended_meta.extended_meta_flagged_rssi import ExtendedMetaFlaggedRssi
from ..utils.extended_meta.extended_meta_flagged_timestamp import ExtendedMetaFlaggedTimestamp
from ..utils.extended_meta.extended_meta_legacy_channel_id import ExtendedMetaLegacyChannelId

logger = logging.getLogger(__name__)


class OpenantPlusControllerHrmFeature(AntplusControllerHrmFeature):
    """
    Setup Level feature implementation for the :class:`AntplusControllerHrmFeature`, by using the
    `openant library <https://github.com/Tigge/openant>`_.
    """

    manager = OpenantManagerFeature()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._openant_channel: Union[Channel, None] = None
        self._broadcast_message_queue = queue.Queue()
        self._ack_message_queue = queue.Queue()
        self._burst_message_queue = queue.Queue()

    @property
    def extended_format(self) -> Literal['legacy', 'flagged', 'none']:
        """
        :return: returns the extended format that should be requested for the ANT channel
        """
        return 'flagged'

    def open_channel(self):
        if self._openant_channel is not None:
            raise ValueError('can not open channel, because another one is still active')

        # reset saved messages TODO
        self._already_saved_broadcast_messages = PageMessageCollection()
        self._already_saved_ack_messages = PageMessageCollection()
        self._already_saved_burst_messages = PageMessageCollection()

        self._openant_channel = self.manager.node.new_channel(self.channel_type, 0x00, 0x01) # TODO configurable?

        # configure callbacks based on if slave or master device

        # configure for MASTER
        self._openant_channel.on_broadcast_data = self._on_broadcast_data
        self._openant_channel.on_burst_data = self._on_burst_data
        self._openant_channel.on_acknowledge_data = self._on_acknowledge
        # only search timeout if slave as searching
        self._openant_channel.set_search_timeout(0xFF)

        self._openant_channel.set_id(self.AntPlusDevice.config.device_num, self.device_type, self.transmission_type)

        self._openant_channel.enable_extended_messages(self.extended_format == 'legacy')

        if self.extended_format == 'legacy':
            self._openant_channel.enable_extended_messages(1)

        if self.extended_format == 'flagged':
            # we will activate everything in flagged mode
            message = Message(Message.ID.LIB_CONFIG, [self._openant_channel.id, 0x80 | 0x20])
            self._openant_channel._ant.write_message(message) # pylint: disable=protected-access

        self._openant_channel.set_period(self.channel_period)
        self._openant_channel.set_rf_freq(self.rf_channel_frequency)

        logger.debug(
            f"opening channel #{self._openant_channel.id}, TYPE 0x{self.channel_type:02x} "
            f"dID {self.AntPlusDevice.config.device_num}; dType {self.device_type}; "
            f"dTrans 0x{self.transmission_type:02x} {self.rf_channel_frequency} @ "
            f"{self.channel_period * 1000 / 0xFFFF:.2f} ms"
        )
        self._openant_channel.open()

    @property
    def channel_is_active(self) -> bool:
        return self._openant_channel and self._openant_channel in self.manager.node.channels

    @property
    def received_broadcast_messages(self) -> PageMessageCollection:
        # read all messages from queue
        msg = self._read_and_save_broadcast_message()

        while msg is not None:
            msg = self._read_and_save_broadcast_message()

        return super().received_broadcast_messages

    @property
    def received_ack_messages(self) -> PageMessageCollection:
        # read all messages from queue
        msg = self._read_and_save_ack_message()

        while msg is not None:
            msg = self._read_and_save_ack_message()

        return super().received_ack_messages

    def get_page_for_no(self, page_no: int) -> type[BaseReceivedAntplusPage]:
        """
        This method returns the page type object for the given page number. It raises a KeyError in case that
        there is no page specified for the given page-number.

        :param page_no: the page number the page type should be returned
        :return: the page type that describes the given page number
        """
        all_hrm_pages = self.AntPlusDevice.profile.get_existing_pages_for_profile()
        if page_no not in all_hrm_pages:
            raise KeyError(f'unable to find page #{page_no}')
        return all_hrm_pages[page_no]

    def _on_broadcast_data(self, data: array.array):
        timestamp = time.perf_counter()
        self._broadcast_message_queue.put((timestamp, data.tobytes()))

    def _on_acknowledge(self, data: array.array):
        timestamp = time.perf_counter()
        self._ack_message_queue.put((timestamp, data.tobytes()))

    def _on_burst_data(self, data: array.array):
        timestamp = time.perf_counter()
        self._burst_message_queue.put((timestamp, data.tobytes()))

    def _get_page_from_raw_data(self, raw_data: bytes) -> type[BaseReceivedAntplusPage]:
        page_no = raw_data[0] & ~(1 << 7)  # on HRM profile -> first bit is toggle bit
        return self.get_page_for_no(page_no)

    @classmethod
    def _parse_legacy_extended_message(
            cls,
            raw_data: bytes
    ) -> list[ExtendedMetaLegacyChannelId]:
        if len(raw_data) != 12:
            raise ValueError(f'expected 12 bytes but for legacy message format {len(raw_data)}')
        return [ExtendedMetaLegacyChannelId(raw_data[0:4])]

    @classmethod
    def _parse_flagged_extended_message(
            cls,
            raw_data: bytes
    ) -> list[Union[ExtendedMetaFlaggedChannelId, ExtendedMetaFlaggedRssi, ExtendedMetaFlaggedTimestamp]]:
        if len(raw_data) <= 8:
            raise ValueError(f'expected more than 8 bytes but for flagged message format {len(raw_data)}')
        extended_data = raw_data[8:]
        all_metas = []
        nxt_idx = 1
        if extended_data[0] & 0x80:
            # next 4 bytes are for `ExtendedMetaLegacyChannelId`
            all_metas.append(ExtendedMetaFlaggedChannelId(extended_data[nxt_idx:nxt_idx + 4]))
            nxt_idx += 4
        if extended_data[0] & 0x40:
            # next 3 bytes are for `ExtendedMetaFlaggedRssi`
            all_metas.append(ExtendedMetaFlaggedRssi(extended_data[nxt_idx:nxt_idx + 3]))
            nxt_idx += 3
        if extended_data[0] & 0x20:
            # next 2 bytes are for `ExtendedMetaFlaggedTimestamp`
            all_metas.append(ExtendedMetaFlaggedTimestamp(extended_data[nxt_idx:nxt_idx + 2]))
            nxt_idx += 2
        if extended_data[nxt_idx:]:
            raise ValueError(f'more extended (flagged) data received than expected: {raw_data}')
        return all_metas

    def _read_from_queue(self, msg_queue: queue.Queue) -> Union[BaseReceivedAntplusPage, None]:
        if msg_queue.empty():
            return None
        timestamp, raw_data = msg_queue.get()

        if self.extended_format == 'none':
            raw_data_of_page_only = raw_data
            meta = None
        elif self.extended_format == 'legacy':
            raw_data_of_page_only = raw_data[4:12]
            meta = self._parse_legacy_extended_message(raw_data)
        elif self.extended_format == 'flagged':
            raw_data_of_page_only = raw_data[0:8]
            meta = self._parse_flagged_extended_message(raw_data)
        else:
            raise ValueError(f'received unexpected value for legacy format `{self.extended_format}`')
        page_type = self._get_page_from_raw_data(raw_data_of_page_only)
        return page_type(raw_data_of_page_only, timestamp=timestamp, extended_metas=meta)

    def _read_and_save_broadcast_message(self) -> Union[BaseAntplusPage, None]:
        msg = self._read_from_queue(self._broadcast_message_queue)
        if msg is None:
            return None
        self._already_saved_broadcast_messages.append(msg)
        return msg

    def _read_and_save_ack_message(self) -> Union[BaseAntplusPage, None]:
        msg = self._read_from_queue(self._ack_message_queue)
        if msg is None:
            return None
        self._already_saved_ack_messages.append(msg)
        return msg

    def send_broadcast_message(self, message: BaseAntplusPage) -> None:
        self._openant_channel.send_broadcast_data(list(message.raw_data))

    def send_ack_message(self, message: BaseAntplusPage):
        self._openant_channel.send_acknowledged_data(list(message.raw_data))

    def close_channel(self) -> bool:
        if self._openant_channel is None:
            return False

        self.manager.node.remove_channel(self._openant_channel)

        # load all messages that are still in queue
        self._read_and_save_broadcast_message()
        self._read_and_save_ack_message()

        self._openant_channel = None
        return True

    def wait_for_new_broadcast_message(
            self,
            of_page_type: Union[list[type[BaseAntplusPage]], type[BaseAntplusPage], None] = None,
            timeout: float = 10
    ) -> BaseAntplusPage:
        # read buffer
        _ = self.received_broadcast_messages

        of_page_type = [of_page_type] if isinstance(of_page_type, type) else of_page_type
        start_time = time.perf_counter()
        while (time.perf_counter() - start_time) < timeout:
            new_msg = self._read_and_save_broadcast_message()
            if new_msg is None:
                time.sleep(0.1)
                continue
            if of_page_type is not None and new_msg.__class__ not in tuple(of_page_type):
                continue
            return new_msg
        raise TimeoutError(f'not received any messages within {timeout} seconds')

    def wait_for_new_ack_message(
            self,
            of_page_type: Union[list[type[BaseAntplusPage]], type[BaseAntplusPage], None] = None,
            timeout: float = 10
    ) -> BaseAntplusPage:
        # read buffer
        _ = self.received_ack_messages

        of_page_type = [of_page_type] if isinstance(of_page_type, type) else of_page_type

        start_time = time.perf_counter()
        while (time.perf_counter() - start_time) < timeout:
            new_msg = self._read_and_save_ack_message()
            if new_msg is None:
                time.sleep(0.1)
                continue
            if of_page_type is not None and new_msg.__class__ not in tuple(of_page_type):
                continue
            return new_msg
        raise TimeoutError(f'not received any messages within {timeout} seconds')



from .base_hrm_scenario import BaseHrmScenario
from .scenario_hrm_battery_messureing import ScenarioHrmBatteryMeasuring
from .scenario_hrm_full_transmission_pattern import ScenarioHrmDeviceProfileFullTransmissionPattern
from .scenario_hrm_manual_request_for_ack import ScenarioManualRequestForAck
from .scenario_hrm_manual_request_for_brdcst import ScenarioHrmManualRequestForBrdcst

__all__ = [
    "BaseHrmScenario",
    "ScenarioHrmBatteryMeasuring",
    "ScenarioHrmDeviceProfileFullTransmissionPattern",
    "ScenarioManualRequestForAck",
    "ScenarioHrmManualRequestForBrdcst",
]



from .base_extended_meta_flagged import BaseExtendedMetaFlagged


class ExtendedMetaFlaggedRssi(BaseExtendedMetaFlagged):
    """
    Metadata describing data given by the ANT interface in FLAGGED-EXTENDED-DATA message format that describes
    additional RSSI information
    """
    EXPECTED_BYTE_LENGTH = 3

    @property
    def measurement_type(self) -> int:
        """
        :return: returns the measurement type
        """
        return self._raw_data[0]

    @property
    def rssi(self) -> int:
        """
        :return: returns the current RSSI value
        """
        return int.from_bytes(self._raw_data[1:1], byteorder=self.BYTE_ORDER, signed=True)

    @property
    def threshold_config_value(self) -> int:
        """
        :return: returns the threshold config value
        """
        return int.from_bytes(self._raw_data[2:2], byteorder=self.BYTE_ORDER, signed=True)



from .page_message_collection import PageMessageCollection
from .support import filter_hrm_messages_by_toggle_bit_change


__all__ = [
    'PageMessageCollection',
    'filter_hrm_messages_by_toggle_bit_change'
]



from .ant_node_manager_feature import AntNodeManagerFeature
from .antplus_controller_feature import AntplusControllerFeature
from .antplus_controller_hrm_feature import AntplusControllerHrmFeature
from .antplus_device_config import AntplusDeviceConfig
from .antplus_hrm_device_config import AntplusHrmDeviceConfig
from .antplus_hrm_test_criteria_config import AntplusHrmTestCriteriaConfig
from .antplus_test_criteria_config import AntplusTestCriteriaConfig
from .base_antplus_device_profile import BaseAntplusDeviceProfile
from .heart_rate_monitor_device_profile import HeartRateMonitorDeviceProfile

__all__ = [
    'AntNodeManagerFeature',
    'AntplusControllerFeature',
    'AntplusControllerHrmFeature',
    'AntplusDeviceConfig',
    'AntplusHrmDeviceConfig',
    'AntplusHrmTestCriteriaConfig',
    'AntplusTestCriteriaConfig',
    'BaseAntplusDeviceProfile',
    'HeartRateMonitorDeviceProfile',
]



import balder

class AntNodeManagerFeature(balder.Feature):
    """
    Base ANT Node Manager Feature - provides bindings to the ANT Stick or the Ant device and manages the connection to
    it. The :class:`AntplusControllerFeature` features use this feature to instantiate their own ANT+ channel.
    """

    @property
    def network_and_network_key(self) -> tuple[int, list[int]]:  # TODO typing
        """
        :return: returns the network and its key the manager should use
        """
        return 0x00, [0xB9, 0xA5, 0x21, 0xFB, 0xBD, 0x72, 0xC3, 0x45]

    def start(self) -> None:
        """
        Starts the manager
        """
        raise NotImplementedError()

    def shutdown(self) -> None:
        """
        Shuts down the manager
        """
        raise NotImplementedError()



import time

import balderhub.heart.lib.scenario_features
from balderhub.ant.lib.scenario_features import AntplusControllerHrmFeature


class RRValueReaderFeature(balderhub.heart.lib.scenario_features.RRValueReaderFeature):
    """
    Setup Level feature implementation of the ``balderhub.heart.lib.scenario_features.RRValueReaderFeature`` while
    reading the values over ANT
    """
    ant_controller = AntplusControllerHrmFeature()

    def read_last_rr_value_in_sec(self) -> float:
        total_timeout = 3
        if not self.ant_controller.channel_is_active:
            self.ant_controller.open_channel()
        start_time = time.perf_counter()
        while time.perf_counter() - start_time < 3:
            cur_msg = self.ant_controller.wait_for_new_broadcast_message(timeout=1)
            if len(set(msg.heart_beat_count for msg in self.ant_controller.received_broadcast_messages)) > 1:
                # more than one messages here -> break
                break
        else:
            raise TimeoutError(f'did not received more than one heart beat messages within {total_timeout} seconds')

        heart_beat_count_before = cur_msg.heart_beat_count

        all_msgs = reversed(self.ant_controller.received_broadcast_messages.messages)
        for msg in all_msgs:
            if msg.heart_beat_count not in [heart_beat_count_before, cur_msg.heart_beat_count]:
                raise ValueError(f'detect message with unexpected heart beat count of {msg.heart_beat_count} '
                                 f'(current message is {cur_msg.heart_beat_count})')
            if msg.heart_beat_count == heart_beat_count_before:
                cur_msg_event_time = cur_msg.heart_beat_event_time
                cur_msg_event_time += 0xFFFF if cur_msg_event_time < msg.heart_beat_event_time else 0

                return (msg.heart_beat_event_time - cur_msg_event_time) / 1024
        raise ValueError('was unable to detect a correct previous heart beat message')



from .base_antplus_device_profile import BaseAntplusDeviceProfile
from ...lib.utils import pages


class HeartRateMonitorDeviceProfile(BaseAntplusDeviceProfile):
    """
    This feature can be assigned to an ANT+ device that supports the Heart Rate Monitor profile.
    """

    ALL_PAGES = [
        pages.hrm.Hrm0DefaultDataPage,
        pages.hrm.Hrm1CumulativeOperationTimePage,
        pages.hrm.Hrm2ManufacturerInformationPage,
        pages.hrm.Hrm3ProductInformationPage,
        pages.hrm.Hrm4PreviousHeartBeatEventTimePage,
        pages.hrm.Hrm5SwimIntervalSummaryPage,
        pages.hrm.Hrm6CapabilitiesPage,
        pages.hrm.Hrm7BatteryStatusPage,
        pages.hrm.Hrm9DeviceInformationPage,
    ]



from __future__ import annotations

from typing import Union

from .base_hrm_page import BaseHrmPage
from ...page_message_collection import PageMessageCollection


class Hrm7BatteryStatusPage(BaseHrmPage):
    """
    This is the seventh data page in the HRM profile that holds information about the battery within the page-specific
    bytes.
    """
    PAGE_ID = 7
    _STRUCT_DATA_FORMAT = '<BBBBHBB'

    @property
    def battery_level(self) -> int:
        """
        :return: returns the battery level
        """
        return self._raw_data[1]

    @property
    def fractional_battery_voltage(self) -> int:
        """
        :return: returns the raw fractional battery voltage if given (use
                 :meth:`Hrm7BatteryStatusPage.total_battery_voltage` for the full voltage information)
        """
        return self._raw_data[2]

    @property
    def descriptive_bit_field(self) -> int:
        """
        :return: returns the full raw descriptive bit field
        """
        return self._raw_data[3]

    @property
    def coarse_battery_voltage(self) -> int:
        """
        :return: returns the coarse battery voltage value
        """
        return self.descriptive_bit_field & 0xF

    @property
    def battery_status(self) -> int:
        """
        :return: returns the battery status value
        """
        return (self.descriptive_bit_field & 0x70) >> 4

    @property
    def total_battery_voltage(self) -> Union[float, None]:
        """
        :return: returns the full battery voltage value or None if it is marked as invalid
        """
        if self.coarse_battery_voltage == 0xF:
            return None
        return (self.descriptive_bit_field & 0x0F) + (self.fractional_battery_voltage / 255)

    @classmethod
    def validate_messages(
            cls,
            of_msg_collection: PageMessageCollection,
            expected_battery_level: int,
            expected_battery_status: Union[int, None],
            expected_battery_voltage: Union[float, None],
            allowed_deviation_for_battery_voltage_percent: float = 0.0
    ) -> None:
        """
        Method to validate the correctness of messages from this type

        Note that with this check all Battery messages within the provided collection need to have the same value.
        Split it by yourself in case you need to validate different values.


        :param of_msg_collection: the message collection that should be validated (will automatically be filtered for
                                  messages from :class:`Hrm7BatteryStatusPage`)
        :param expected_battery_level: the expected battery level that should be in all messages of this type
        :param expected_battery_status: the expected battery status that should be in all messages of this type or None,
                                        if it is expected that the device sends INVALID (0xF) here
        :param expected_battery_voltage: the expected battery voltage that should be in all messages of this type or
                                         None, if it is expected that the device sends INVALID (0xF in coarse voltage)
                                         here
        """
        relevant_msgs = of_msg_collection.filter_by_type(page_type=cls)

        vals_for_battery_level = list(relevant_msgs.get_unique_values_for_field('battery_level'))
        assert len(vals_for_battery_level) == 1 and vals_for_battery_level[0] == expected_battery_level, \
            (f"detect unexpected values for `battery_level` for {cls.__name__} messages: `{vals_for_battery_level}` "
             f"(expected {expected_battery_level})")
        assert 0 <= vals_for_battery_level[0] <= 100, \
            f"battery level is not in expected range of 0-100, is {vals_for_battery_level[0]}"

        if expected_battery_voltage:
            vals_for_bat_voltage = list(relevant_msgs.get_unique_values_for_field('total_battery_voltage'))
            assert len(vals_for_bat_voltage) == 1, \
                (f"detect unexpected count of values for `total_battery_voltage` for {cls.__name__} messages: "
                 f"`{vals_for_bat_voltage}`")

            min_expected_voltage = expected_battery_voltage * (1 - allowed_deviation_for_battery_voltage_percent)
            max_expected_voltage = expected_battery_voltage * (1 + allowed_deviation_for_battery_voltage_percent)
            assert min_expected_voltage <= vals_for_bat_voltage[0] <= max_expected_voltage, \
                (f"battery voltage of {vals_for_bat_voltage[0]} is not in expected range "
                 f"(expected voltage is {expected_battery_voltage} "
                 f"+/- {allowed_deviation_for_battery_voltage_percent*100:.2f}%: "
                 f"{min_expected_voltage:.3f}V - {max_expected_voltage:.3f}V)")

            val_for_coarse_voltage = list(relevant_msgs.get_unique_values_for_field('coarse_battery_voltage'))
            assert len(val_for_coarse_voltage) == 1, (f"detect unexpected values for `coarse_battery_voltage` for "
                                                      f"{cls.__name__} messages: `{vals_for_bat_voltage}`")
            assert 0 < val_for_coarse_voltage[0] < 0xF, \
                f"coarse_battery_voltage has invalid value {val_for_coarse_voltage}"
        else:
            # battery voltage is expected to be set to INVALID
            fractional_battery_voltage = relevant_msgs.get_unique_value_for_field('fractional_battery_voltage')
            assert fractional_battery_voltage == 0xFF, ("fractional battery voltage should be set to 0xFF, because "
                                                        "it's expected that the battery voltage is transmitted as "
                                                        "INVALID")

            coarse_voltage = relevant_msgs.get_unique_value_for_field('coarse_battery_voltage')
            assert coarse_voltage == 0xF, ("coarse battery voltage within descriptive bit field should be set to 0xF, "
                                           "because it's expected that the battery voltage is transmitted as INVALID")

        val_for_bat_status = relevant_msgs.get_unique_value_for_field('battery_status')
        if expected_battery_status:
            assert 0x00 < val_for_bat_status < 0x07, \
                f"battery status for message has invalid value {val_for_bat_status}"

            assert val_for_bat_status == expected_battery_status, \
                (f"detect unexpected values for `battery_status` for {cls.__name__} messages: `{val_for_bat_status}`"
                 f" (expected {expected_battery_status})")

        else:
            assert val_for_bat_status == 0x07, \
                f"battery status for message does not have expected invalid value {val_for_bat_status}"



from __future__ import annotations
from .base_hrm_page import BaseHrmPage
from ...page_message_collection import PageMessageCollection


class Hrm3ProductInformationPage(BaseHrmPage):
    """
    This is the second data page in the HRM profile that holds some product information within the page-specific
    bytes.
    """
    PAGE_ID = 3
    _STRUCT_DATA_FORMAT = '<BBBBHBB'

    @property
    def hardware_version(self) -> int:
        """
        :return: returns the hardware version of the device
        """
        return self.raw_data[1]

    @property
    def software_version(self) -> int:
        """
        :return: returns the software version of the device
        """
        return self.raw_data[2]

    @property
    def model_number(self) -> int:
        """
        :return: returns the model number of the device
        """
        return self.raw_data[3]

    @classmethod
    def validate_messages(
            cls,
            of_msg_collection: PageMessageCollection,
            expected_hardware_version: int,
            expected_software_version: int,
            expected_model_number: int
    ) -> None:
        """
        Method to validate the correctness of messages from this type

        :param of_msg_collection: the message collection that should be validated (will automatically be filtered for
                                  messages from :class:`Hrm3ProductInformationPage`)
        :param expected_hardware_version: the expected hardware version that should be in all messages of this type
        :param expected_software_version: the expected software version that should be in all messages of this type
        :param expected_model_number: the expected model number that should be in all messages of this type
        :return:
        """
        relevant_msgs = of_msg_collection.filter_by_type(cls)

        vals_for_hwversion = relevant_msgs.get_unique_values_for_field('hardware_version')
        assert len(vals_for_hwversion) == 1 and list(vals_for_hwversion)[0] == expected_hardware_version, \
            f"detect unexpected values for `hardware_version` for {cls.__name__} messages: `{vals_for_hwversion}`"

        vals_for_swversion = relevant_msgs.get_unique_values_for_field('software_version')
        assert len(vals_for_swversion) == 1 and list(vals_for_swversion)[0] == expected_software_version, \
            f"detect unexpected values for `software_version` for {cls.__name__} messages: `{vals_for_swversion}`"

        vals_for_model_number = relevant_msgs.get_unique_values_for_field('model_number')
        assert len(vals_for_model_number) == 1 and list(vals_for_model_number)[0] == expected_model_number, \
            f"detect unexpected values for `model_number` for {cls.__name__} messages: `{vals_for_model_number}`"



from .base_antplus_page import BaseAntplusPage
from .base_received_antplus_page import BaseReceivedAntplusPage
from . import common, hrm

__all__ = [
    'BaseAntplusPage',
    'BaseReceivedAntplusPage',
]



import time
from datetime import timedelta, datetime
import logging

import balder
from balder.connections import DCPowerConnection

from balderhub.ant.lib.utils import pages
import balderhub.battery.lib.scenario_features
from balderhub.heart.lib.scenario_features import HeartBeatFeature, StrapDockingFeature

from .base_hrm_scenario import BaseHrmScenario
from ...lib.utils.page_message_collection import PageMessageCollection

logger = logging.getLogger(__name__)


class ScenarioManualRequestForAck(BaseHrmScenario):
    """
    Test scenario for validating the correct responses when sending a :class:`Common70RequestDataPage` and ask for ACK
    messages as response. This scenario tests every possible background page and expects that the HR-Sensor responses
    with an ACK message as answer. It also validates its content.

    In case the page is not mentioned within :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`, the tests
    within this scenario expect that the DUT does not respond with the requested page. Instead, it just continues its
    usual messages without sending anything as ACK.

    If :meth:`AntplusHrmDeviceConfig.manual_request_redirect_ack_as_broadcast` returns true every test within this
    scenario expects, that the DUT responds with a BROADCAST message even when requested for an ACK message (like it
    is done with every page-request within this scenario).
    """

    # TODO further test ideas: wait till we expect exactly the background page and then request it (-> new scenario)
    # TODO further test idea: request a page (transmission-count = 100) and directly after that ask for another one
    # TODO also check the illegal value for request-transmission-response: 0x00 (problems???)
    # TODO can I request another page after I've tried to request an invalid Common Page?

    DO_WITH_BATTERY_LEVEL = 1.0
    ADDITIONAL_WAIT_SEC = 5

    class Heart(balder.Device):
        """device simulating the heart beat"""
        heart = HeartBeatFeature()

    @balder.connect('HeartRateSensor', over_connection=DCPowerConnection)
    class BatterySimulator(balder.Device):
        """device manipulating the battery voltage"""
        sim = balderhub.battery.lib.scenario_features.RemovableBatterySimFeature()

    @balder.connect(Heart, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateSensor(BaseHrmScenario.HeartRateSensor):
        """device detecting the row heart rate"""
        # _is_cnn = IsConnected()  # TODO
        strap = StrapDockingFeature()

    @balder.connect(HeartRateSensor, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateHost(BaseHrmScenario.HeartRateHost):
        """device receiving the heart rate data"""

    @balder.fixture('variation')
    def device_powered_on(self):
        """fixture that ensures that the device is powered on, before entering the variation"""
        yield from self.BatterySimulator.sim.fixt_make_sure_device_is_powered_on(
            with_level=self.DO_WITH_BATTERY_LEVEL,
            restore_entry_state=True
        )

    @balder.fixture('variation')
    def heart_beat_established(self, device_powered_on):  # pylint: disable=unused-argument
        """fixture that ensures that the heart beat is active, before entering the variation"""
        yield from self.Heart.heart.fixt_make_sure_heart_beat_established(with_bpm=60, restore_entry_state=True)

    @balder.fixture('variation')
    def chest_strap_attached(self, heart_beat_established):  # pylint: disable=unused-argument
        """fixture that ensures that the chest strap is attached, before entering the variation"""
        yield from self.HeartRateSensor.strap.fixt_make_sure_to_be_attached(restore_entry_state=True)

    @balder.fixture('variation')
    def ant_connected(self, chest_strap_attached):  # pylint: disable=unused-argument
        """fixture that ensures that ANT channel is open, before entering the variation"""
        yield from self.HeartRateHost.controller.fixt_make_sure_ant_channel_is_opened()

    @classmethod
    def get_page_to_send(
            cls,
            page_to_request: type[pages.hrm.BaseHrmPage],
            transmit_no: int
    ) -> pages.common.Common70RequestDataPage:
        """
        Helper method to get the Request-Data-Page object filled with the given data

        :param page_to_request: the page that should be requested
        :param transmit_no: the requested transmit number
        :return: the filled and ready to send request-page
        """
        assert 0 < transmit_no <= 0x7F, f"illegal value for transmit_no: {transmit_no}"
        return pages.common.Common70RequestDataPage.create(0x80 | transmit_no, page_to_request.PAGE_ID, 0x01)

    def _do_request_for_ack(
            self,
            page_to_request: type[pages.hrm.BaseHrmPage],
            transmit_no: int
    ) -> PageMessageCollection:
        page_to_send = self.__class__.get_page_to_send(page_to_request=page_to_request, transmit_no=transmit_no)

        # do sync and wait for the first background page to be fully transmitted - does not matter which background page
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_background_pages,
            timeout=20)  # increase timeout because 64 / 4 ~ 16 second
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_main_page)
        # after background page was fully transmitted -> send request (now it should normally return main pages)
        timestamp_before = datetime.now()
        self.HeartRateHost.controller.send_ack_message(page_to_send)
        # TODO be careful if we wait more than 16 seconds -> then the expectation would be to receive a background page
        # wait for all messages to be transmitted (with some additional seconds - we want to check that it continues
        #  with main pages after that)
        time.sleep(max(1, transmit_no) * 0.25 + self.ADDITIONAL_WAIT_SEC)

        relevant_brdcst_msgs = self.HeartRateHost.controller.received_broadcast_messages.filter_for_timestamp(
            start=timestamp_before)

        relevant_ack_msgs = self.HeartRateHost.controller.received_ack_messages.filter_for_timestamp(
            start=timestamp_before)
        assert len(relevant_ack_msgs) == len(relevant_ack_msgs.filter_by_type(page_type=page_to_request)), \
            f"unexpectedly received ACK messages from another type than {page_to_request.__name__}: {relevant_ack_msgs}"
        assert len(relevant_ack_msgs) == max(transmit_no, 1), \
            (f"received unexpected message count - requested {transmit_no} times over Request page, "
             f"but received {len(relevant_ack_msgs)} messages of requested type")

        assert (relevant_ack_msgs[0].timestamp - timestamp_before) < timedelta(seconds=1), \
            "response time is higher than expected"  # TODO make configurable

        # make sure that we only received main pages afterwards
        brdcst_msgs_of_requested_page_type = relevant_brdcst_msgs.filter_by_type(page_type=page_to_request)
        assert len(brdcst_msgs_of_requested_page_type) == 0, \
            (f"received some broadcast messages from requested type {page_to_request.__name__} - unexpected because we "
             f"ask for ACK messages")

        assert len(relevant_brdcst_msgs) > 0, "no other broadcast messages has been sent while asking for ACK messages"

        brdcst_msg_types = relevant_brdcst_msgs.get_message_types()

        assert (len(brdcst_msg_types) == 1
                and list(brdcst_msg_types)[0] == self.HeartRateSensor.ant_config.expected_main_page), \
            (f"received messages from unexpected type (only main pages were expected after all responses for Page 70 "
             f"Request have been transmitted): {brdcst_msgs_of_requested_page_type}")
        return relevant_ack_msgs

    def _do_request_for_ack_but_expect_brdcst_response(
            self,
            page_to_request: type[pages.hrm.BaseHrmPage],
            transmit_no: int
    ) -> PageMessageCollection:
        page_to_send = self.__class__.get_page_to_send(page_to_request=page_to_request, transmit_no=transmit_no)

        # do sync and wait for the first background page to be fully transmitted - does not matter which background page
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_background_pages,
            timeout=20)  # increase timeout because 64 / 4 ~ 16 second
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_main_page)
        # after background page was fully transmitted -> send request (now it should normally return main pages)
        timestamp_before = datetime.now()
        self.HeartRateHost.controller.send_ack_message(page_to_send)
        # TODO be careful if we wait more than 16 seconds -> then the expectation would be to receive a background page
        # wait for all messages to be transmitted (with some additional seconds - we want to check that it continues
        #  with main pages after that)
        time.sleep(max(1, transmit_no) * 0.25 + self.ADDITIONAL_WAIT_SEC)

        relevant_brdcst_msgs = self.HeartRateHost.controller.received_broadcast_messages.filter_for_timestamp(
            start=timestamp_before)

        # make sure that we did not receive any ACK messages
        relevant_ack_msgs = self.HeartRateHost.controller.received_ack_messages.filter_for_timestamp(
            start=timestamp_before
        )
        assert len(relevant_ack_msgs) == 0, \
            (f"received unexpected ACK messages while requesting page {page_to_request.__name__} by ACK, but while "
             f"expecting that test return messages as broadcast (setting in "
             f"`{self.HeartRateSensor.ant_config.__class__}.manual_request_redirect_ack_as_broadcast`): "
             f"{relevant_ack_msgs}")

        # TODO this needs to cleaned up for message-loss
        # make sure that we received exactly the transmit-no count of messages
        msgs_of_requested_page_type = relevant_brdcst_msgs.filter_by_type(page_type=page_to_request)

        assert len(msgs_of_requested_page_type) == transmit_no, \
            (f"received unexpected count of messages - received {len(msgs_of_requested_page_type)} messages of "
             f"type {page_to_request.__name__}, but expected length was {transmit_no}")

        first_msg = msgs_of_requested_page_type[0]
        last_msg = msgs_of_requested_page_type[-1]

        # check Request-Page Message response time (time between request and first response)
        # TODO this needs to cleaned up for message-loss
        assert (first_msg.timestamp - timestamp_before) < timedelta(seconds=1), \
            "response time is higher than expected"  # TODO make configurable

        # make sure that we only received main pages afterwards
        remaining_msgs = relevant_brdcst_msgs.filter_for_timestamp(start=last_msg.timestamp + timedelta(seconds=0.1))

        assert len(remaining_msgs) > 0, ("expect some more main page messages after last transferred requested page "
                                         "messages, but did not receive anything more")

        message_types = list(remaining_msgs.get_message_types())
        assert (len(message_types) == 1 and message_types[0] == self.HeartRateSensor.ant_config.expected_main_page), \
            (f"received messages from unexpected type (only main pages were expected after all responses for Page 70 "
             f"Request have been transmitted): {remaining_msgs}")
        return msgs_of_requested_page_type

    def _do_request_for_ack_but_expect_no_response(
            self,
            page_to_request: type[pages.hrm.BaseHrmPage],
            transmit_no: int
    ) -> None:
        page_to_send = self.__class__.get_page_to_send(page_to_request=page_to_request, transmit_no=transmit_no)

        # do sync and wait for the first background page to be fully transmitted - does not matter which background page
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_background_pages,
            timeout=20  # increase timeout because 64 / 4 ~ 16 second
        )
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_main_page
        )

        # after background page was fully transmitted -> send request (now it should normally return main pages)
        timestamp_before = datetime.now()
        self.HeartRateHost.controller.send_ack_message(page_to_send)
        # wait for all messages to be transmitted (with some additional seconds - we want to make sure that no pages
        # has been transmitted)
        time.sleep(max(1, transmit_no) * 0.25 + self.ADDITIONAL_WAIT_SEC)
        # make sure that we did not receive any ACK messages
        relevant_ack_msgs = self.HeartRateHost.controller.received_ack_messages.filter_for_timestamp(
            start=timestamp_before
        )

        assert len(relevant_ack_msgs) == 0, \
            (f"received unexpected ACK messages while requesting a non-active page {page_to_request.__name__}: "
             f"{relevant_ack_msgs}")

        relevant_brdcst_msgs = self.HeartRateHost.controller.received_broadcast_messages.filter_for_timestamp(
            start=timestamp_before
        )

        # make sure that we have only received main pages
        msg_types = relevant_brdcst_msgs.get_message_types()
        assert len(msg_types) == 1 and list(msg_types)[0] == self.HeartRateSensor.ant_config.expected_main_page, \
            (f"received unexpected broadcast messages while requesting a non-active page "
             f"{page_to_request.__name__}: {msg_types}")

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_ack')
    )
    def test_ack_page_1_operating_time(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a ACK message of
        :class:`Hrm1CumulativeOperationTimePage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If :meth:`AntplusHrmDeviceConfig.manual_request_redirect_ack_as_broadcast` is True, the test validates that
        the DUT responds with a BROADCAST message instead of responding with the requested ACK message.

        Depending on what is expected, the test also checks that no messages from requested type arrive as ACK or
        BROADCAST if this is not expected.

        If the test expects a response it will also validate the content of the messages.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm1CumulativeOperationTimePage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            if self.HeartRateSensor.ant_config.manual_request_redirect_ack_as_broadcast:
                response = self._do_request_for_ack_but_expect_brdcst_response(page_type, transmit_no)
            else:
                response = self._do_request_for_ack(page_type, transmit_no)

            page_type.validate_messages(
                    of_msg_collection=response,
            )
        else:
            self._do_request_for_ack_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_ack')
    )
    def test_ack_page_2_manufacturer(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a ACK message of
        :class:`Hrm2ManufacturerInformationPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If :meth:`AntplusHrmDeviceConfig.manual_request_redirect_ack_as_broadcast` is True, the test validates that
        the DUT responds with a BROADCAST message instead of responding with the requested ACK message.

        Depending on what is expected, the test also checks that no messages from requested type arrive as ACK or
        BROADCAST if this is not expected.

        If the test expects a response it will also validate the content of the messages.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm2ManufacturerInformationPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            if self.HeartRateSensor.ant_config.manual_request_redirect_ack_as_broadcast:
                response = self._do_request_for_ack_but_expect_brdcst_response(page_type, transmit_no)
            else:
                response = self._do_request_for_ack(page_type, transmit_no)

            page_type.validate_messages(
                of_msg_collection=response,
                expected_manufacturer_id=self.HeartRateSensor.ant_config.manufacturer_id,
                expected_serial_number=self.HeartRateSensor.ant_config.serial_number
            )

        else:
            self._do_request_for_ack_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_ack')
    )
    def test_ack_page_3_product(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a ACK message of
        :class:`Hrm3ProductInformationPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If :meth:`AntplusHrmDeviceConfig.manual_request_redirect_ack_as_broadcast` is True, the test validates that
        the DUT responds with a BROADCAST message instead of responding with the requested ACK message.

        Depending on what is expected, the test also checks that no messages from requested type arrive as ACK or
        BROADCAST if this is not expected.

        If the test expects a response it will also validate the content of the messages.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm3ProductInformationPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            if self.HeartRateSensor.ant_config.manual_request_redirect_ack_as_broadcast:
                response = self._do_request_for_ack_but_expect_brdcst_response(page_type, transmit_no)
            else:
                response = self._do_request_for_ack(page_type, transmit_no)

            page_type.validate_messages(
                of_msg_collection=response,
                expected_hardware_version=self.HeartRateSensor.ant_config.hardware_version,
                expected_software_version=self.HeartRateSensor.ant_config.software_version,
                expected_model_number=self.HeartRateSensor.ant_config.model_number,
            )
        else:
            self._do_request_for_ack_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_ack')
    )
    def test_ack_page_6_capabilities(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a ACK message of
        :class:`Hrm6CapabilitiesPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If :meth:`AntplusHrmDeviceConfig.manual_request_redirect_ack_as_broadcast` is True, the test validates that
        the DUT responds with a BROADCAST message instead of responding with the requested ACK message.

        Depending on what is expected, the test also checks that no messages from requested type arrive as ACK or
        BROADCAST if this is not expected.

        If the test expects a response it will also validate the content of the messages.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm6CapabilitiesPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            if self.HeartRateSensor.ant_config.manual_request_redirect_ack_as_broadcast:
                __response = self._do_request_for_ack_but_expect_brdcst_response(page_type, transmit_no)
            else:
                __response = self._do_request_for_ack(page_type, transmit_no)

                # TODO validate message content
                raise NotImplementedError()
        else:
            self._do_request_for_ack_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_ack')
    )
    def test_ack_page_7_battery(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a ACK message of
        :class:`Hrm7BatteryStatusPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If :meth:`AntplusHrmDeviceConfig.manual_request_redirect_ack_as_broadcast` is True, the test validates that
        the DUT responds with a BROADCAST message instead of responding with the requested ACK message.

        Depending on what is expected, the test also checks that no messages from requested type arrive as ACK or
        BROADCAST if this is not expected.

        If the test expects a response it will also validate the content of the messages.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm7BatteryStatusPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            if self.HeartRateSensor.ant_config.manual_request_redirect_ack_as_broadcast:
                response = self._do_request_for_ack_but_expect_brdcst_response(page_type, transmit_no)
            else:
                response = self._do_request_for_ack(page_type, transmit_no)

            expected_bat_voltage = \
                self.BatterySimulator.sim.discharge_characteristic.get_voltage_for(self.DO_WITH_BATTERY_LEVEL) \
                    if self.HeartRateSensor.ant_config.support_battery_voltage_messuring else None

            page_type.validate_messages(
                of_msg_collection=response,
                expected_battery_level=int(self.DO_WITH_BATTERY_LEVEL * 100),
                expected_battery_status=self.HeartRateSensor.ant_config.get_expected_battery_state_for_level(
                    int(self.DO_WITH_BATTERY_LEVEL * 100)
                ),
                expected_battery_voltage=expected_bat_voltage,
            )
        else:
            self._do_request_for_ack_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_ack')
    )
    def test_ack_page_9_device_info(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a ACK message of
        :class:`Hrm9DeviceInformationPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If :meth:`AntplusHrmDeviceConfig.manual_request_redirect_ack_as_broadcast` is True, the test validates that
        the DUT responds with a BROADCAST message instead of responding with the requested ACK message.

        Depending on what is expected, the test also checks that no messages from requested type arrive as ACK or
        BROADCAST if this is not expected.

        If the test expects a response it will also validate the content of the messages.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm9DeviceInformationPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            if self.HeartRateSensor.ant_config.manual_request_redirect_ack_as_broadcast:
                response = self._do_request_for_ack_but_expect_brdcst_response(page_type, transmit_no)
            else:
                response = self._do_request_for_ack(page_type, transmit_no)

            page_type.validate_messages(
                of_msg_collection=response,
            )
        else:
            self._do_request_for_ack_but_expect_no_response(page_type, transmit_no)



import balder


class AntplusDeviceConfig(balder.Feature):
    """Base universal ANT+ Device configuration feature"""

    @property
    def device_num(self):
        """
        :return: returns the device number of the device
        """
        raise NotImplementedError()



from typing import Literal, Any
from abc import ABC

import struct


class BaseAntplusPage(ABC):
    """Base Abstract class describing an ANT+ page"""
    PAGE_ID = None # needs to be defined in final page
    _STRUCT_DATA_FORMAT = None
    BYTEORDER: Literal['big', 'little'] = 'little'

    def __init__(
            self,
            raw_data: bytes,
    ):

        if not isinstance(self.__class__.PAGE_ID, int):
            raise TypeError('PAGE_ID must be an integer between 0 and 255')
        if not 0x00 < self.__class__.PAGE_ID <= 0xFF:
            raise ValueError('PAGE_ID must be between 0 and 255')

        if not isinstance(self.__class__._STRUCT_DATA_FORMAT, str):
            raise TypeError('STRUCT_DATA_FORMAT must be a string')

        self._validate_raw_data(raw_data)
        self._validate_page_num(raw_data)

        self._raw_data = raw_data

    def __repr__(self):
        return f"{self.__class__.__name__}<{self._raw_data}>"

    @property
    def raw_data(self) -> bytes:
        """
        :return: the raw data of this page
        """
        return self._raw_data

    def _validate_raw_data(self, for_raw_data: bytes) -> None:
        if not isinstance(for_raw_data, bytes):
            raise TypeError(
                f'raw data must be a bytes object, but is from type {for_raw_data.__class__.__name__}: {for_raw_data}'
            )
        if len(for_raw_data) != 8:
            raise ValueError(f'raw data must be of length 8, but is {len(for_raw_data)} bytes')

    def _validate_page_num(self, in_raw_data: bytes) -> None:
        if in_raw_data[0] != self.PAGE_ID:
            raise ValueError(f'raw data holds wrong page number {in_raw_data[0]} '
                             f'(expected {self.PAGE_ID} for {self.__class__.__name__})')

    def _raw_unpack(self) -> tuple[Any, ...]:
        return struct.unpack(self._STRUCT_DATA_FORMAT, self.raw_data)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.raw_data == other.raw_data

    def __hash__(self):
        return hash(self.raw_data) + hash(self.__class__)



from balderhub.ant.lib.utils.page_message_collection import PageMessageCollection
from balderhub.ant.lib.utils.pages.base_received_antplus_page import BaseReceivedAntplusPage


def filter_hrm_messages_by_toggle_bit_change(
    messages: PageMessageCollection,
) -> PageMessageCollection:
    """
    Helper function to filter all messages from a given page-message collection to only return the according to the
    spec relevant messages, because the toggle bit has changed.

    :param messages: the message collection that should be filtered
    :return: a new message collection that only holds the first messages after every toggle bit change
    """
    result = PageMessageCollection()

    if len(messages) == 0:
        return result

    def get_toggle_bit(msg: BaseReceivedAntplusPage) -> bool:
        return bool(msg.raw_data[0] & 0x80)

    toggle_bit_before = not get_toggle_bit(messages[0])
    for msg in messages:
        cur_toggle_bit = get_toggle_bit(msg)
        if toggle_bit_before != cur_toggle_bit:
            result.append(msg)
            toggle_bit_before = cur_toggle_bit
    return result



import logging
import threading
from typing import Union

from openant.easy.node import Node

from ..scenario_features.ant_node_manager_feature import AntNodeManagerFeature

logger = logging.getLogger(__name__)


class OpenantManagerFeature(AntNodeManagerFeature):
    """
    Setup Level feature implementation of :class:`balderhub.ant.lib.scenario_features.AntNodeManagerFeature` that uses
    the `OpenAnt Python Library <https://github.com/Tigge/openant>`_ to interact with remote devices
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._thread = None
        self._node = None

    @property
    def node(self) -> Union[Node, None]:
        """
        :return: returns the ``openant.easy.node.Node`` object
        """
        if self._node is None:
            raise ValueError('manager needs to start service before node can be accessed')
        return self._node

    def _threaded_method(self):
        logger.debug('openant manager thread started.')
        self._node.start()

    def start(self):
        self._thread = threading.Thread(target=self._threaded_method)

        self._node = Node()
        self._node.set_network_key(*self.network_and_network_key)

        self._thread.start()

    def shutdown(self, timeout=5) -> bool:
        if self._node is None:
            return False
        self.node.stop()
        self._thread.join(timeout=timeout)
        if self._thread.is_alive():
            raise RuntimeError('manager thread failed to shut down')
        self._node = None
        self._thread = None
        return True



from __future__ import annotations
from .base_hrm_page import BaseHrmPage


class Hrm5SwimIntervalSummaryPage(BaseHrmPage):
    """
    This is the fifth data page in the HRM profile that holds information about swiming intervals within the
    page-specific bytes.
    """
    PAGE_ID = 5
    _STRUCT_DATA_FORMAT = '<BBBBHBB'

    @property
    def interval_average_heart_rate(self):
        """
        :return: returns the average heart rate within the interval
        """
        return self.raw_data[1]

    @property
    def interval_maximum_heart_rate(self):
        """
        :return: returns the maximum heart rate within the interval
        """
        return self.raw_data[2]

    @property
    def session_average_heart_rate(self):
        """
        :return: returns the average heart rate within the session
        """
        return self.raw_data[3]



from __future__ import annotations
from ..base_antplus_page import BaseAntplusPage


class Hrm32HrFeaturePage(BaseAntplusPage):
    """page for controlling HR features"""
    PAGE_ID = 32
    _STRUCT_DATA_FORMAT = '<BBBBBBBB'



from .base_extended_meta_flagged import BaseExtendedMetaFlagged


class ExtendedMetaLegacyChannelId(BaseExtendedMetaFlagged):
    """
    Metadata describing data given by the ANT interface in LEGACY-EXTENDED-DATA message format that describes
    additional Channel ID information
    """
    EXPECTED_BYTE_LENGTH = 4

    @property
    def device_number(self) -> int:
        """
        :return: returns the device number
        """
        return int.from_bytes(self._raw_data[0:1], byteorder=self.BYTE_ORDER, signed=False)

    @property
    def device_type(self) -> int:
        """
        :return: returns the device type
        """
        return self._raw_data[2]

    @property
    def trans_type(self) -> int:
        """
        :return: returns the transport type
        """
        return self._raw_data[3]



import time
from datetime import datetime, timedelta
import logging

import balder
from balder.connections import DCPowerConnection

from balderhub.ant.lib.utils import pages
import balderhub.battery.lib.scenario_features
from balderhub.heart.lib.scenario_features import HeartBeatFeature, StrapDockingFeature

from .base_hrm_scenario import BaseHrmScenario
from ...lib.utils.page_message_collection import PageMessageCollection

logger = logging.getLogger(__name__)


class ScenarioHrmManualRequestForBrdcst(BaseHrmScenario):
    """
    Test scenario for validating the correct responses when sending a :class:`Common70RequestDataPage` and ask for
    BROADCAST messages as response. This scenario tests every possible background page and expects that the HR-Sensor
    responses with an BROADCAST message as answer. It also validates its content.

    In case the page is not mentioned within :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`, the tests
    within this scenario expect that the DUT does not respond with the requested page. Instead, it just continues its
    usual messages without sending anything as ACK.
    """

    # TODO further test ideas: wait till we expect exactly the background page and then request it (new scenario)
    # TODO further test idea: request a page (transmission-count = 100) and directly after that ask for another one
    # TODO also check the illegal value for request-transmission-response: 0x00 (problems?)
    # TODO it would be also interesting if we do the test before the normal timeslot of a background page
    # TODO can I request another page after I've tried to request an invalid Common Page?

    DO_WITH_BATTERY_LEVEL = 1.0
    ADDITIONAL_WAIT_SEC = 5


    class Heart(balder.Device):
        """heart beat simulating device"""
        heart = HeartBeatFeature()

    @balder.connect('HeartRateSensor', over_connection=DCPowerConnection)
    class BatterySimulator(balder.Device):
        """device manipulating the battery voltage"""
        sim = balderhub.battery.lib.scenario_features.RemovableBatterySimFeature()

    @balder.connect(Heart, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateSensor(BaseHrmScenario.HeartRateSensor):
        """device detecting the row heart rate"""
        strap = StrapDockingFeature()

    @balder.connect(HeartRateSensor, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateHost(BaseHrmScenario.HeartRateHost):
        """device receiving the heart rate data"""

    @balder.fixture('variation')
    def device_powered_on(self):
        """fixture that ensures that the device is powered on, before entering the variation"""
        yield from self.BatterySimulator.sim.fixt_make_sure_device_is_powered_on(
            with_level=self.DO_WITH_BATTERY_LEVEL,
            restore_entry_state=True
        )

    @balder.fixture('variation')
    def heart_beat_established(self, device_powered_on):  # pylint: disable=unused-argument
        """fixture that ensures that the heart beat is active, before entering the variation"""
        yield from self.Heart.heart.fixt_make_sure_heart_beat_established(with_bpm=60, restore_entry_state=True)

    @balder.fixture('variation')
    def chest_strap_attached(self, heart_beat_established):  # pylint: disable=unused-argument
        """fixture that ensures that the chest strap is attached, before entering the variation"""
        yield from self.HeartRateSensor.strap.fixt_make_sure_to_be_attached(restore_entry_state=True)

    @balder.fixture('variation')
    def ant_connected(self, chest_strap_attached):  # pylint: disable=unused-argument
        """fixture that ensures that ANT channel is open, before entering the variation"""
        yield from self.HeartRateHost.controller.fixt_make_sure_ant_channel_is_opened()

    @classmethod
    def get_page_to_send(
            cls,
            page_to_request: type[pages.hrm.BaseHrmPage],
            transmit_no: int
    ) -> pages.common.Common70RequestDataPage:
        """
        Helper method to get the Request-Data-Page object filled with the given data

        :param page_to_request: the page that should be requested
        :param transmit_no: the requested transmit number
        :return: the filled and ready to send request-page
        """
        assert 0 < transmit_no <= 0x7F, f"illegal value for transmit_no: {transmit_no}"
        return pages.common.Common70RequestDataPage.create(0x00 | transmit_no, page_to_request.PAGE_ID, 0x01)

    def _do_request_for_brdcst(
            self,
            page_to_request: type[pages.hrm.BaseHrmPage],
            transmit_no: int
    ) -> PageMessageCollection:
        page_to_send = self.__class__.get_page_to_send(page_to_request=page_to_request, transmit_no=transmit_no)

        # do sync and wait for the first background page to be fully transmitted - does not matter which background page
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_background_pages,
            timeout=20  # increase timeout because 64 / 4 ~ 16 second
        )
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_main_page
        )
        # after background page was fully transmitted -> send request (now it should normally return main pages)
        timestamp_before = datetime.now()
        self.HeartRateHost.controller.send_broadcast_message(page_to_send)

        # wait for all messages to be transmitted (with some additional seconds - we want to check that it continues
        #  with main pages after that)
        time.sleep(max(1, transmit_no) * 0.25 + self.ADDITIONAL_WAIT_SEC)
        # make sure that we did not receive any ACK messages
        relevant_ack_msgs = self.HeartRateHost.controller.received_ack_messages.filter_for_timestamp(
            start=timestamp_before
        )
        assert len(relevant_ack_msgs) == 0, \
            f"received some unexpected ACK messages during timeslot: {relevant_ack_msgs}"

        relevant_brdcst_msgs = self.HeartRateHost.controller.received_broadcast_messages.filter_for_timestamp(
            start=timestamp_before
        )

        # make sure that we received exactly the transmit-no count of messages
        # TODO this needs to cleaned up for message-loss
        msgs_of_requested_page_type = relevant_brdcst_msgs.filter_by_type(page_type=page_to_request)

        assert len(msgs_of_requested_page_type) == transmit_no, \
            (f"received unexpected count of messages - received {len(msgs_of_requested_page_type)} messages of "
             f"type {page_to_request.__name__}, but expected length was {transmit_no}")

        first_msg = msgs_of_requested_page_type[0]
        last_msg = msgs_of_requested_page_type[-1]

        # check Request-Page Message response time (time between request and first response)
        # TODO this needs to cleaned up for message-loss
        assert (first_msg.timestamp - timestamp_before) < timedelta(seconds=1), \
            "response time is higher than expected" # TODO make configurable


        # make sure that we only received main pages afterwards
        remaining_msgs = relevant_brdcst_msgs.filter_for_timestamp(start=last_msg.timestamp + timedelta(seconds=0.1))

        assert len(remaining_msgs) > 0, ("expect some more main page messages after last transferred requested page "
                                         "messages, but did not receive anything more")

        all_msg_types = list(remaining_msgs.get_message_types())
        assert (len(all_msg_types) == 1 and all_msg_types[0] == self.HeartRateSensor.ant_config.expected_main_page), \
            (f"received messages from unexpected type (only main pages were expected after all responses for Page 70 "
             f"Request have been transmitted): {remaining_msgs}")

        return msgs_of_requested_page_type

    def _do_request_for_brdcst_but_expect_no_response(
            self,
            page_to_request: type[pages.hrm.BaseHrmPage],
            transmit_no: int
    ) -> None:
        page_to_send = self.__class__.get_page_to_send(page_to_request=page_to_request, transmit_no=transmit_no)

        # do sync and wait for the first background page to be fully transmitted - does not matter which background page
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_background_pages,
            timeout=20  # increase timeout because 64 / 4 ~ 16 second
        )
        self.HeartRateHost.controller.wait_for_new_broadcast_message(
            of_page_type=self.HeartRateSensor.ant_config.expected_main_page
        )

        # after background page was fully transmitted -> send request (now it should normally return main pages)
        timestamp_before = datetime.now()
        self.HeartRateHost.controller.send_broadcast_message(page_to_send)
        # wait for all messages to be transmitted (with some additional seconds - we want to make sure that no pages
        # has been transmitted)
        time.sleep(max(1, transmit_no) * 0.25 + self.ADDITIONAL_WAIT_SEC)
        # make sure that we did not receive any ACK messages
        relevant_ack_msgs = self.HeartRateHost.controller.received_ack_messages.filter_for_timestamp(
            start=timestamp_before
        )
        assert len(relevant_ack_msgs) == 0, \
            f"received some unexpected ACK messages during timeslot: {relevant_ack_msgs}"

        relevant_brdcst_msgs = self.HeartRateHost.controller.received_broadcast_messages.filter_for_timestamp(
            start=timestamp_before
        )

        # make sure that we have only received main pages
        msg_types = relevant_brdcst_msgs.get_message_types()
        assert len(msg_types) == 1 and list(msg_types)[0] == self.HeartRateSensor.ant_config.expected_main_page, \
            (f"received unexpected broadcast messages while requesting a non-active page "
             f"{page_to_request.__name__}: {msg_types}")

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_broadcast')
    )
    def test_brdcst_page_1_operating_time(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a BROADCAST message of
        :class:`Hrm1CumulativeOperationTimePage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If the test expects a response it will also validate the content of the messages. Additionally, it makes sure
        that no messages are sent as ACK because it requests BROADCAST messages only.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm1CumulativeOperationTimePage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            response = self._do_request_for_brdcst(page_type, transmit_no)

            page_type.validate_messages(
                of_msg_collection=response,
            )
        else:
            self._do_request_for_brdcst_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_broadcast')
    )
    def test_brdcst_page_2_manufacturer(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a BROADCAST message of
        :class:`Hrm2ManufacturerInformationPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If the test expects a response it will also validate the content of the messages. Additionally, it makes sure
        that no messages are sent as ACK because it requests BROADCAST messages only.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm2ManufacturerInformationPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            response = self._do_request_for_brdcst(page_type, transmit_no)

            page_type.validate_messages(
                of_msg_collection=response,
                expected_manufacturer_id=self.HeartRateSensor.ant_config.manufacturer_id,
                expected_serial_number=self.HeartRateSensor.ant_config.serial_number
            )
        else:
            self._do_request_for_brdcst_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_broadcast')
    )
    def test_brdcst_page_3_product(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a BROADCAST message of
        :class:`Hrm3ProductInformationPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If the test expects a response it will also validate the content of the messages. Additionally, it makes sure
        that no messages are sent as ACK because it requests BROADCAST messages only.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm3ProductInformationPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            response = self._do_request_for_brdcst(page_type, transmit_no)

            page_type.validate_messages(
                of_msg_collection=response,
                expected_hardware_version=self.HeartRateSensor.ant_config.hardware_version,
                expected_software_version=self.HeartRateSensor.ant_config.software_version,
                expected_model_number=self.HeartRateSensor.ant_config.model_number,
            )
        else:
            self._do_request_for_brdcst_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_broadcast')
    )
    def test_brdcst_page_6_capabilities(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a BROADCAST message of
        :class:`Hrm6CapabilitiesPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If the test expects a response it will also validate the content of the messages. Additionally, it makes sure
        that no messages are sent as ACK because it requests BROADCAST messages only.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm6CapabilitiesPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:  # pylint: disable=no-else-raise
            __response = self._do_request_for_brdcst(page_type, transmit_no)
            raise NotImplementedError('this page is not fully implemented yet')
            # TODO validate message content
        else:
            self._do_request_for_brdcst_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_broadcast')
    )
    def test_brdcst_page_7_battery(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a BROADCAST message of
        :class:`Hrm7BatteryStatusPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If the test expects a response it will also validate the content of the messages. Additionally, it makes sure
        that no messages are sent as ACK because it requests BROADCAST messages only.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        page_type = pages.hrm.Hrm7BatteryStatusPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            response = self._do_request_for_brdcst(page_type, transmit_no)

            expected_bat_voltage = \
                (self.BatterySimulator.sim.discharge_characteristic.get_voltage_for(self.DO_WITH_BATTERY_LEVEL)
                 if self.HeartRateSensor.ant_config.support_battery_voltage_messuring else None)

            page_type.validate_messages(
                of_msg_collection=response,
                expected_battery_level=int(self.DO_WITH_BATTERY_LEVEL * 100),
                expected_battery_status=self.HeartRateSensor.ant_config.get_expected_battery_state_for_level(
                    int(self.DO_WITH_BATTERY_LEVEL * 100)
                ),
                expected_battery_voltage=expected_bat_voltage,
            )
        else:
            self._do_request_for_brdcst_but_expect_no_response(page_type, transmit_no)

    @balder.parametrize_by_feature(
        'transmit_no', (HeartRateSensor, 'test_criteria', 'request_transmission_numbers_for_broadcast')
    )
    def test_brdcst_page_9_device_info(self, transmit_no: int):
        """
        Test that validates the correct behavior, when the controller ask for a BROADCAST message of
        :class:`Hrm9DeviceInformationPage`. The test validates that the DUT answers the correct count of requested
        messages, when the page is mentioned in :meth:`AntplusHrmDeviceConfig.manual_request_possible_for`. Otherwise,
        it will ensure that the DUT ignores the request.

        If the test expects a response it will also validate the content of the messages. Additionally, it makes sure
        that no messages are sent as ACK because it requests BROADCAST messages only.

        :param transmit_no: PARAMETRIZED value describing the requested times the message should be sent
        """
        # check that it is not within the transmission or if data is valid (depending on setting)
        page_type = pages.hrm.Hrm9DeviceInformationPage

        if page_type in self.HeartRateSensor.ant_config.manual_request_possible_for:
            response = self._do_request_for_brdcst(page_type, transmit_no)

            page_type.validate_messages(
                of_msg_collection=response,
            )
        else:
            self._do_request_for_brdcst_but_expect_no_response(page_type, transmit_no)



from __future__ import annotations

from .base_hrm_page import BaseHrmPage
from ...page_message_collection import PageMessageCollection


class Hrm2ManufacturerInformationPage(BaseHrmPage):
    """
    This is the second data page in the HRM profile that holds some manufacturer information within the page-specific
    bytes.
    """
    PAGE_ID = 2
    _STRUCT_DATA_FORMAT = '<BBHHBB'

    @property
    def manufacturer_id(self) -> int:
        """
        :return: returns the manufacturer id of the device
        """
        return self.raw_data[1]

    @property
    def serial_number(self) -> int:
        """
        :return: holds the upper 16 bits of the serial number of the device
        """
        return int.from_bytes(self.raw_data[2:3], 'little')

    @classmethod
    def validate_messages(
            cls,
            of_msg_collection: PageMessageCollection,
            expected_manufacturer_id: int,
            expected_serial_number: int
    ) -> None:
        """
        Method to validate the correctness of messages from this type

        :param of_msg_collection: the message collection that should be validated (will automatically be filtered for
                                  messages from :class:`Hrm2ManufacturerInformationPage`)
        :param expected_manufacturer_id: the expected manufacturer id that should be in all messages of this type
        :param expected_serial_number: the expected serial number that should be in all messages of this type
        """
        relevant_msgs = of_msg_collection.filter_by_type(cls)

        vals_for_manufacturer_id = relevant_msgs.get_unique_values_for_field('manufacturer_id')
        assert len(vals_for_manufacturer_id) == 1 and list(vals_for_manufacturer_id)[0] == expected_manufacturer_id, \
            f"detect unexpected values for `manufacturer_id` for {cls.__name__} messages: `{vals_for_manufacturer_id}`"

        vals_for_serial_number = relevant_msgs.get_unique_values_for_field('serial_number')
        assert len(vals_for_serial_number) == 1 and list(vals_for_serial_number)[0] == expected_serial_number, \
            f"detect unexpected values for `serial_number` for {cls.__name__} messages: `{vals_for_serial_number}`"



import balder


class AntplusTestCriteriaConfig(balder.Feature):
    """
    General ANT+ test criteria configuration feature
    """

    @property
    def request_transmission_numbers_for_broadcast(self) -> list[int]:
        """
        :return: returns a list of all transmission-numbers that should be tried an validated within
                 :class:`ScenarioHrmManualRequestForBrdcst`
        """
        return [1, 5]

    @property
    def request_transmission_numbers_for_ack(self) -> list[int]:
        """
        :return: returns a list of all transmission-numbers that should be tried an validated within
                 :class:`ScenarioHrmManualRequestForAck`
        """
        return [1, 5]

    @property
    def allowed_packet_loss_percent(self) -> float:
        """value between 0 and 1 that defines the accepted package loss during transmission"""
        return 0

    @property
    def first_number_of_beats_to_skip(self): # TODO improve name
        """
        :return: indicates the number of beat that are ignored before the cyclic values like RR-Value, BPM,
                 consecutive beat counts, ... are validated
        """
        return 0



from __future__ import annotations

from datetime import timedelta

from .base_hrm_page import BaseHrmPage
from ...page_message_collection import PageMessageCollection


class Hrm1CumulativeOperationTimePage(BaseHrmPage):
    """
    This is the first data page in the HRM profile that holds the cumulative operation time within the page-specific
    bytes.
    """
    PAGE_ID = 1
    _STRUCT_DATA_FORMAT = '<BBBBHBB'

    @property
    def cumulative_operating_time_raw(self) -> int:
        """
        :return: returns the raw cumulative operation time (increments every two seconds)
        """
        return int.from_bytes(self.raw_data[1:4], 'little')

    @property
    def cumulative_operating_time_sec(self) -> int:
        """
        :return: returns the cumulative operation time in seconds
        """
        return self.cumulative_operating_time_raw * 2

    @classmethod
    def validate_messages(
            cls,
            of_msg_collection: PageMessageCollection,
    ) -> None:
        """
        Method to validate the correctness of messages from this type

        :param of_msg_collection: the message collection that should be validated (will automatically be filtered for
                                  messages from :class:`Hrm1CumulativeOperationTimePage`)
        """
        relevant_msgs = of_msg_collection.filter_by_type(page_type=cls)

        # TODO improve validation algorithm
        # counter should increase every two seconds
        if len(relevant_msgs) == 0:
            raise ValueError(f'did not receive any messages for {cls.__name__}')
        assert (relevant_msgs[-1].timestamp - relevant_msgs[0].timestamp) <= timedelta(seconds=2.25), \
            "did not get enough messages to be able to validate Operation-Time counter"

        cleared_optimes = [msg.cumulative_operating_time_raw for msg in relevant_msgs]
        idx_before = 0
        for idx in range(1, len(cleared_optimes)):
            if cleared_optimes[idx_before] <= cleared_optimes[idx]:
                # no overflow -> continue
                continue
            # increase value for all following items
            for sub_idx in range(idx, len(cleared_optimes)):
                cleared_optimes[sub_idx] += 0xFFFFFF
            idx_before = idx

        # determine synchron time
        sync_timestamp = relevant_msgs[0].timestamp
        sync_optime = relevant_msgs[0].cumulative_operating_time_raw

        for idx in range(1, len(relevant_msgs)):
            msg = relevant_msgs[idx]
            diff_timestamp = msg.timestamp - sync_timestamp

            diff_optime_raw = cleared_optimes[idx] - sync_optime
            diff_optime_sec = diff_optime_raw * 2

            # check if difftime is the same (accuracy needs to be around 2 seconds)
            assert diff_optime_sec - 2 < diff_timestamp < diff_optime_sec + 2, \
                f"detect illegal times since message {msg}"



from __future__ import annotations

from abc import ABC

from balderhub.ant.lib.utils.pages.base_received_antplus_page import BaseReceivedAntplusPage


class BaseHrmPage(BaseReceivedAntplusPage, ABC):
    """
    Base Heart-Rate-Monitor Page
    """
    PAGE_ID = None

    def __repr__(self):
        return (f"{self.__class__.__name__}<toggle: {int(self.toggle_bit)} "
                f"| event-time={self.heart_beat_event_time} "
                f"| beat-count={self.heart_beat_count} "
                f"| beat={self.computed_heart_rate} ({self._timestamp:.4f})>")

    def _validate_page_num(self, in_raw_data: bytes) -> None:
        """custom page num validation page -> allows toggle bit"""
        toggle_bit_adjusted_page_num = in_raw_data[0] & ~(1 << 7)
        if toggle_bit_adjusted_page_num != self.PAGE_ID:
            raise ValueError(f'raw data holds wrong page number {toggle_bit_adjusted_page_num} (toggle bit adjusted)'
                             f'(expected {self.PAGE_ID} for {self.__class__.__name__})')

    @property
    def toggle_bit(self) -> bool:
        """
        :return: the current state of the toggle bit of the current page message
        """
        unpacked_data = self._raw_unpack()
        return bool((unpacked_data[0] >> 7) & 0x01)

    @property
    def heart_beat_event_time(self) -> int:
        """
        :return: the raw heart-beat-event-time of the current page message
        """
        unpacked_data = self._raw_unpack()
        return unpacked_data[-3]

    @property
    def heart_beat_count(self) -> int:
        """
        :return: the raw heart-beat-count of the current page message
        """
        unpacked_data = self._raw_unpack()

        return unpacked_data[-2]

    @property
    def computed_heart_rate(self) -> int:
        """
        :return: the computed-heart-rate of the current page message
        """
        unpacked_data = self._raw_unpack()
        return unpacked_data[-1]



import time
from typing import Optional, Union, TypeVar

from abc import ABC
from datetime import datetime, timedelta

from balderhub.ant.lib.utils.extended_meta.base_extended_meta_legacy import BaseExtendedMetaLegacy
from balderhub.ant.lib.utils.extended_meta.base_extended_meta_flagged import BaseExtendedMetaFlagged
from .base_antplus_page import BaseAntplusPage

BaseReceivedAntplusPageTypeT = TypeVar('BaseReceivedAntplusPageTypeT', bound='BaseReceivedAntplusPage')


class BaseReceivedAntplusPage(BaseAntplusPage, ABC):
    """
    Base ANT+ Page that is intended to be received from a remote ANT+ device (holds a valid timestamp and metadata)
    """

    def __init__(
            self,
            raw_data: bytes,
            timestamp: float,
            extended_metas: Optional[Union[list[BaseExtendedMetaLegacy], list[BaseExtendedMetaFlagged]]] = None,
    ):

        super().__init__(raw_data=raw_data)

        self._extended_metas = []

        if extended_metas:
            base_class = BaseExtendedMetaFlagged \
                if isinstance(extended_metas[0], BaseExtendedMetaFlagged) \
                else BaseExtendedMetaLegacy

            for meta in extended_metas:
                if not isinstance(meta, base_class):
                    raise TypeError(
                        f'meta objects needs to be either all subclasses of {BaseExtendedMetaLegacy.__name__} '
                        f'or of type {BaseExtendedMetaFlagged.__name__}, '
                    )
                meta_types = [meta.__class__ for meta in extended_metas]
                if len(meta_types) != len(set(meta_types)):
                    raise ValueError('not provide multiple meta objects of same type')
                self._extended_metas.append(meta)

        self._timestamp = timestamp


    def __repr__(self):
        return f"{self.__class__.__name__}<{self._timestamp:.4f}: {self._raw_data}>"

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        return self._timestamp == other._timestamp

    def __hash__(self):
        return hash(self.raw_data) + hash(self.__class__) + hash(self._timestamp)

    @property
    def timestamp(self) -> datetime:
        """
        .. note::
            This is not ment to be a timestamp determined by the ANT host directly. This timestamp is created as soon as
            the message receives the balderhub python implementation.

        :return: returns the timestamp the message was received by the feature
        """
        return datetime.now() - timedelta(seconds=time.perf_counter() - self._timestamp)



from __future__ import annotations

import sys
from typing import Iterator, SupportsIndex, Union, Any, TYPE_CHECKING

from datetime import datetime
if TYPE_CHECKING:
    from .pages.base_received_antplus_page import BaseReceivedAntplusPage, BaseReceivedAntplusPageTypeT


class PageMessageCollection:
    """
    Page Message Collection object to manage multiple page messages of the same or different type
    """

    def __init__(self, initial_messages: list[BaseReceivedAntplusPageTypeT] = None):
        self._messages = []

        if initial_messages:
            for msg in initial_messages:
                self.append(msg)

    def __repr__(self):
        return f"{self.__class__.__name__}<{str(self._messages)}>"

    def __iter__(self) -> Iterator[BaseReceivedAntplusPageTypeT]:
        return iter(self._messages)

    def __len__(self):
        return len(self._messages)

    def __getitem__(self, item: int) -> BaseReceivedAntplusPageTypeT:
        return self._messages[item]

    @property
    def messages(self) -> list[BaseReceivedAntplusPageTypeT]:
        """
        :return: returns a copy of all internal messages
        """
        return self._messages.copy()

    def append(self, message: BaseReceivedAntplusPageTypeT) -> None:
        """
        Adds a message to the collection. It will be automatically inserted in the correct order according to its
        timestamp.

        :param message: the message that should be added to the collection
        """
        from .pages.base_received_antplus_page import BaseReceivedAntplusPage  # pylint: disable=import-outside-toplevel

        if not isinstance(message, BaseReceivedAntplusPage):
            raise TypeError(f'messages need to be a subclass of type {BaseReceivedAntplusPage}')

        self._messages.append(message)

        self._sort_by_timestamp()

    def index(
            self,
            value: BaseReceivedAntplusPageTypeT,
            start: SupportsIndex = 0,
            stop: SupportsIndex = sys.maxsize
    ) -> int:
        """
        Returns the index of the next (similar to ``list.index``)

        :param value: the value to look for
        :param start: first index to look at
        :param stop: last index to look at
        :return: the index within the internal messsage list
        """
        return self._messages.index(value, start, stop)

    def _sort_by_timestamp(self):
        self._messages.sort(key=lambda x: x.timestamp)

    def filter_by_type(
            self,
            page_type: type[BaseReceivedAntplusPageTypeT]
    ) -> PageMessageCollection[BaseReceivedAntplusPageTypeT]:
        """
        Returns a new collection that holds all messages from the given type

        :param page_type: the page type to filter this message collection
        :return: a new collection that holds all messages from the given type
        """
        result = self.__class__()
        for msg in self._messages:
            if msg.__class__ == page_type:
                result.append(msg)
        return result

    def get_unique_values_for_field(
            self,
            field_name: str,
            ignore_non_existing=False
    ) -> set[Any]:
        """
        This method returns a set of values the messages within this collection has. If ``ignore_non_existing`` is False
        the method will raise an exception if there is a message type that does not provide the ``field_name`` as
        property.

        :param field_name: the field name the unique value set should be returned for
        :param ignore_non_existing: True if all messages that does not have this field are ignored, False otherwise
                                    (raises an error)
        :return: the unique set of values that exist within this collection
        """
        all_values = set()
        for msg in self._messages:
            if ignore_non_existing:
                if hasattr(msg, field_name):
                    all_values.add(getattr(msg, field_name))
            else:
                all_values.add(getattr(msg, field_name))
        return all_values

    def get_unique_value_for_field(self, field_name: str) -> Any | None:
        """
        This method returns the unique value for one field. It throws an error if the method finds more than one values
        or if the field does not exist within one or more messages.
        :param field_name:
        :return:
        """
        all_unique_vals = self.get_unique_values_for_field(field_name)
        if len(all_unique_vals) == 1:
            return all_unique_vals.pop()
        raise ValueError('found multiple values for field name `{}` within messages')

    def get_first_message(
            self,
            of_page_type: Union[type[BaseReceivedAntplusPageTypeT], None]
    ) -> Union[BaseReceivedAntplusPageTypeT, None]:
        """
        This method returns the first occurrence of the given page-type within this collection.

        :param of_page_type: the page type to look for
        :return: the first page-message of the requested type that is within this collection or None if no message of
                 this type was found
        """
        filtered = self if of_page_type is None else self.filter_by_type(page_type=of_page_type)
        return filtered[0]

    def get_last_message(
            self,
            of_page_type: Union[type[BaseReceivedAntplusPageTypeT], None]
    ) -> Union[BaseReceivedAntplusPageTypeT, None]:
        """
        This method returns the last occurrence of the given page-type within this collection.

        :param of_page_type: the page type to look for
        :return: the last page-message of the requested type that is within this collection or None if no message of
                 this type was found
        """
        filtered = self if of_page_type is None else self.filter_by_type(page_type=of_page_type)
        return filtered[-1]

    def get_message_types(self) -> set[type[BaseReceivedAntplusPageTypeT]]:
        """
        :return: returns a set object of all page-message types that exist within this collection
        """
        return set(msg.__class__ for msg in self._messages)

    def filter_for_timestamp(
            self,
            start: datetime = None,
            end: datetime = None
    ) -> PageMessageCollection[BaseReceivedAntplusPageTypeT]:
        """
        This method returns all messages which timestamp is between the given start and/or end date
        :param start: start (exclusive) timestamp, the messages can have to be returned
        :param end: end (exclusive) timestamp, the messages can have to be returned
        :return: a new collection that matches the given filter criteria
        """
        result = self.__class__()

        for msg in self._messages:
            if start is not None and msg.timestamp < start:
                continue
            if end is not None and msg.timestamp > end:
                continue
            result.append(msg)
        return result





# TODO according to HRM Device Profile section 5.2 we need to make sure that the master channel is configured correctly

# TODO make sure that the device-number is not 0x0000!! (see secion 5.2 - table row `Device Number`)

# TODO validate that sensor toggles the toggle bit every four messages (validation of profile itself?)



from .base_extended_meta_flagged import BaseExtendedMetaFlagged
from .base_extended_meta_legacy import BaseExtendedMetaLegacy
from .extended_meta_flagged_channel_id import ExtendedMetaFlaggedChannelId
from .extended_meta_flagged_rssi import ExtendedMetaFlaggedRssi
from .extended_meta_flagged_timestamp import ExtendedMetaFlaggedTimestamp
from .extended_meta_legacy_channel_id import ExtendedMetaLegacyChannelId


__all__ = [
    'BaseExtendedMetaFlagged',
    'BaseExtendedMetaLegacy',
    'ExtendedMetaFlaggedChannelId',
    'ExtendedMetaFlaggedRssi',
    'ExtendedMetaFlaggedTimestamp',
    'ExtendedMetaLegacyChannelId',
]



from .base_extended_meta_flagged import BaseExtendedMetaFlagged


class ExtendedMetaFlaggedTimestamp(BaseExtendedMetaFlagged):
    """
    Metadata describing data given by the ANT interface in FLAGGED-EXTENDED-DATA message format that describes
    additional TIMESTAMP information
    """
    EXPECTED_BYTE_LENGTH = 2

    @property
    def timestamp_raw(self) -> int:
        """
        :return: returns the raw timestamp value of the message
        """
        return int.from_bytes(self._raw_data, byteorder=self.BYTE_ORDER, signed=False)

    @property
    def timestamp_sec(self) -> float:
        """
        :return: returns the timestamp in seconds (rolls over every two second)
        """
        return self.timestamp_raw / 32768



import time
from typing import Union


import balderhub.ant.lib.scenario_features
from balderhub.ant.lib.utils import pages
import balderhub.battery.lib.scenario_features


class BatteryLevelReader(balderhub.battery.lib.scenario_features.BatteryLevelReader):
    """Setup Level implementation for reading the battery level over ANT"""
    controller = balderhub.ant.lib.scenario_features.AntplusControllerHrmFeature()

    @property
    def initial_wait_sec(self):
        """
        :return: returns time in seconds to wait for the device to have done the measurement
        """
        return 10

    def read_current_battery_level(self) -> Union[float, None]:
        page = pages.hrm.Hrm7BatteryStatusPage
        request = pages.common.Common70RequestDataPage.create(0x00 | 0x0F, page.PAGE_ID, 0x01)

        channel_was_active_before = self.controller.channel_is_active
        try:
            if not channel_was_active_before:
                self.controller.open_channel()
                time.sleep(self.initial_wait_sec)
            self.controller.send_broadcast_message(request)
            try:
                new_message = self.controller.wait_for_new_broadcast_message(page)
                return new_message.battery_level / 100
            except TimeoutError:
                return None
        finally:
            if not channel_was_active_before:
                self.controller.close_channel()



import logging

import balder

from balderhub.ant.lib.scenario_features.antplus_controller_hrm_feature import AntplusControllerHrmFeature
from balderhub.ant.lib.scenario_features.antplus_hrm_device_config import AntplusHrmDeviceConfig
from balderhub.ant.lib.scenario_features.antplus_hrm_test_criteria_config import AntplusHrmTestCriteriaConfig
from balderhub.ant.lib.scenario_features.heart_rate_monitor_device_profile import HeartRateMonitorDeviceProfile

logger = logging.getLogger(__name__)


class BaseHrmScenario(balder.Scenario):
    """Base test scenario for working with Heart-Rate Monitor devices"""

    class HeartRateSensor(balder.Device):
        """device detecting the row heart rate"""
        ant_config = AntplusHrmDeviceConfig()
        test_criteria = AntplusHrmTestCriteriaConfig()
        hrm = HeartRateMonitorDeviceProfile()

    @balder.connect(HeartRateSensor, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateHost(balder.Device):
        """device receiving the heart rate data"""
        controller = AntplusControllerHrmFeature(AntPlusDevice='HeartRateSensor')
